# Copyright 2018 Goldman Sachs.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
#
# Chart Service will attempt to make public functions (not prefixed with _) from this module available. Such functions
# should be fully documented: docstrings should describe parameters and the return value, and provide a 1-line
# description. Type annotations should be provided for parameters.
import math
from enum import IntEnum, Enum
from typing import Union

import numpy as np
import pandas as pd

from gs_quant.api.gs.data import GsDataApi, QueryType
from gs_quant.common import Currency
from gs_quant.markets.securities import Asset
from .analysis import LagMode, lag
from .datetime import align, interpolate
from .helper import CurveType, Interpolate, Window, normalize_window, apply_ramp, plot_session_function, \
    plot_function, Returns, SeriesType
from .statistics import std, product, sum_
from ..data import DataContext
from ..datetime import DayCountConvention, day_count_fraction
from ..errors import MqValueError, MqTypeError

"""
Econometrics timeseries library is for standard economic and time series analytics operations, including returns,
diffs, lags, volatilities and other numerical operations which are generally finance-oriented
"""


class AnnualizationFactor(IntEnum):
    DAILY = 252
    WEEKLY = 52
    SEMI_MONTHLY = 26
    MONTHLY = 12
    QUARTERLY = 4
    ANNUALLY = 1


class SharpeAssets(Enum):
    USD = 'MAP35DA6K5B1YXGX'
    AUD = 'MAFRZWJ790MQY0EW'
    CHF = 'MAS0NN4ZX7NYXB36'
    EUR = 'MA95W0N1214395N8'
    GBP = 'MA41ZEFTWR8Q7HBM'
    JPY = 'MA8GXV3SJ0TXH1JV'
    SEK = 'MAGNZZY0GJ4TATNG'


def excess_returns_pure(price_series: pd.Series, spot_curve: pd.Series) -> pd.Series:
    curve, bench_curve = align(price_series, spot_curve, Interpolate.INTERSECT)

    e_returns = [curve.iloc[0]]
    for i in range(1, len(curve)):
        multiplier = 1 + curve.iloc[i] / curve.iloc[i - 1] - bench_curve.iloc[i] / bench_curve.iloc[i - 1]
        e_returns.append(e_returns[-1] * multiplier)
    return pd.Series(e_returns, index=curve.index)


def excess_returns(price_series: pd.Series, benchmark_or_rate: Union[Asset, Currency, float], *,
                   day_count_convention=DayCountConvention.ACTUAL_360) -> pd.Series:
    if isinstance(benchmark_or_rate, float):
        er = [price_series.iloc[0]]
        for j in range(1, len(price_series)):
            fraction = day_count_fraction(price_series.index[j - 1], price_series.index[j], day_count_convention)
            er.append(er[-1] + price_series.iloc[j] - price_series.iloc[j - 1] * (1 + benchmark_or_rate * fraction))
        return pd.Series(er, index=price_series.index)

    if isinstance(benchmark_or_rate, Currency):
        try:
            marquee_id = SharpeAssets[benchmark_or_rate.value].value
        except KeyError:
            raise MqValueError(f"unsupported currency {benchmark_or_rate}")
    else:
        marquee_id = benchmark_or_rate.get_marquee_id()

    with DataContext(price_series.index[0], price_series.index[-1]):
        q = GsDataApi.build_market_data_query([marquee_id], QueryType.SPOT)
        df = GsDataApi.get_market_data(q)
    if df.empty:
        raise MqValueError(f'could not retrieve risk-free rate {marquee_id}')
    df = df[~df.index.duplicated(keep='first')]  # handle bad data (duplicate rows)

    return excess_returns_pure(price_series, df['spot'])


def _annualized_return(levels: pd.Series, rolling: Union[int, pd.DateOffset],
                       interpolation_method: Interpolate = Interpolate.NAN) -> pd.Series:
    if isinstance(rolling, pd.DateOffset):
        starting = [tstamp - rolling for tstamp in levels.index]
        levels = interpolate(levels, method=interpolation_method)
        points = list(
            map(lambda d, v, i: pow(v / levels.get(i, np.nan),
                                    365.25 / (d - i).days) - 1,
                levels.index[1:],
                levels.values[1:], starting[1:]))
    else:
        if interpolation_method is not Interpolate.NAN:
            raise MqValueError(f'If w is not a relative date, method must be nan. You specified method: '
                               f'{interpolation_method.value}.')
        starting = [0] * rolling
        starting.extend([a for a in range(1, len(levels) - rolling + 1)])
        points = list(
            map(lambda d, v, i: pow(v / levels.iloc[i], 365.25 / (d - levels.index[i]).days) - 1, levels.index[1:],
                levels.values[1:], starting[1:]))
    points.insert(0, 0)
    return pd.Series(points, index=levels.index)


def get_ratio_pure(er: pd.Series, w: Union[Window, int, str],
                   interpolation_method: Interpolate = Interpolate.NAN) -> pd.Series:
    w = normalize_window(er, w or None)  # continue to support 0 as an input for window
    ann_return = _annualized_return(er, w.w, interpolation_method=interpolation_method)
    long_enough = (er.index[-1] - w.w) >= er.index[0] if isinstance(w.w, pd.DateOffset) else w.w < len(er)
    ann_vol = volatility(er, w).iloc[1:] if long_enough else volatility(er)
    result = ann_return / ann_vol * 100
    return apply_ramp(result, w)


def _get_ratio(input_series: pd.Series, benchmark_or_rate: Union[Asset, float, str], w: Union[Window, int, str], *,
               day_count_convention: DayCountConvention, curve_type: CurveType = CurveType.PRICES,
               interpolation_method: Interpolate = Interpolate.NAN) -> pd.Series:
    if curve_type == CurveType.PRICES:
        er = excess_returns(input_series, benchmark_or_rate, day_count_convention=day_count_convention)
    else:
        assert curve_type == CurveType.EXCESS_RETURNS
        er = input_series

    return get_ratio_pure(er, w, interpolation_method)


class RiskFreeRateCurrency(Enum):
    USD = "USD"
    AUD = "AUD"
    CHF = "CHF"
    EUR = "EUR"
    GBP = "GBP"
    JPY = "JPY"
    SEK = "SEK"
    _USD = "usd"
    _AUD = "aud"
    _CHF = "chf"
    _EUR = "eur"
    _GBP = "gbp"
    _JPY = "jpy"
    _SEK = "sek"


@plot_session_function
def excess_returns_(price_series: pd.Series, currency: RiskFreeRateCurrency = RiskFreeRateCurrency.USD) -> pd.Series:
    """
    Calculate excess returns

    :param price_series: price series
    :param currency: currency for risk-free rate, defaults to USD
    :return: excess returns

    **Usage**

    Given a price series P and risk-free rate R, excess returns E are defined as:

    :math:`E_t = E_{t-1} + P_t - P_{t-1} * (1 + R * (D_t - D_{t-1}) / 360)`

    The `Actual/360 <https://en.wikipedia.org/wiki/Day_count_convention#Actual/360>`_ day count convention is used.

    **Examples**

    Get excess returns from a price series.

    >>> er = excess_returns(generate_series(100), USD)
    """
    return excess_returns(price_series, Currency(currency.value), day_count_convention=DayCountConvention.ACTUAL_360)


@plot_session_function
def sharpe_ratio(series: pd.Series, currency: RiskFreeRateCurrency = RiskFreeRateCurrency.USD,
                 w: Union[Window, int, str] = None, curve_type: CurveType = CurveType.PRICES,
                 method: Interpolate = Interpolate.NAN) -> pd.Series:
    """
    Calculate Sharpe ratio

    :param series: series of prices or excess returns for an asset
    :param currency: currency for risk-free rate, defaults to USD
    :param w: Window or int: size of window and ramp up to use. e.g. Window(22, 10) where 22 is the window size
              and 10 the ramp up value.  If w is a string, it should be a relative date like '1m', '1d', etc.
              Window size defaults to length of series.
    :param curve_type: whether input series is of prices or excess returns, defaults to prices
    :param method: interpolation method (default: nan). Used to calculate returns on dates without data (i.e. weekends)
              when window is a relative date. Defaults to no interpolation.
    :return: Sharpe ratio

    **Usage**

    Given a price series P, risk-free rate R, and window of size w returns the rolling
    `Sharpe ratio <https://en.wikipedia.org/wiki/Sharpe_ratio>`_ S:

    :math:`S_t = \\frac{(E_t / E_{t-w+1})^{365.25 / (D_t - D_{t-w})}-1}{volatility(E, w)_t}`

    Excess returns E are defined as:

    :math:`E_t = E_{t-1} + P_t - P_{t-1} * (1 + R * (D_t - D_{t-1}) / 360)`

    where D is the date for a data point. The
    `Actual/360 <https://en.wikipedia.org/wiki/Day_count_convention#Actual/360>`_ day count convention is used.

    **Examples**

    Get rolling sharpe ratio of a price series (with window of 22).

    >>> sr = sharpe_ratio(generate_series(365, END_TODAY), USD, 22, PRICES)

    **See also**

    :func:`volatility`
    """
    return _get_ratio(series, Currency(currency.value), w, day_count_convention=DayCountConvention.ACTUAL_360,
                      curve_type=curve_type, interpolation_method=method)


@plot_function
def returns(series: pd.Series, obs: Union[Window, int, str] = 1, type: Returns = Returns.SIMPLE) -> pd.Series:
    """
    Calculate returns from price series

    :param series: time series of prices
    :param obs: number of observations or relative date e.g. 3d, 1w, 1m
    :param type: returns type: simple, logarithmic or absolute
    :return: date-based time series of return

    **Usage**

    Compute returns series from price levels, based on the value of *type*:

    ===========   =============================
    Type          Description
    ===========   =============================
    simple        Simple arithmetic returns
    logarithmic   Logarithmic returns
    absolute      Absolute returns
    ===========   =============================

    *Simple*

    Simple geometric change in asset prices, which can be aggregated across assets

    :math:`Y_t = \\frac{X_t}{X_{t-obs}} - 1`

    where :math:`X_t` is the asset price at time :math:`t`

    *Logarithmic*

    Natural logarithm of asset price changes, which can be aggregated through time

    :math:`Y_t = log(X_t) - log(X_{t-obs})`

    where :math:`X_t` is the asset price at time :math:`t`

    *Absolute*

    Absolute change in asset prices

    :math:`Y_t = X_t - X_{t-obs}`

    where :math:`X_t` is the asset price at time :math:`t`

    **Examples**

    Generate price series and take compute returns

    >>> prices = generate_series(100)
    >>> returns = returns(prices)

    **See also**

    :func:`prices`
    """

    if series.size < 1:
        return series

    shifted_series = lag(series, obs, LagMode.TRUNCATE)

    if type == Returns.SIMPLE:
        ret_series = series / shifted_series - 1
    elif type == Returns.LOGARITHMIC:
        ret_series = series.apply(math.log) - shifted_series.apply(math.log)
    elif type == Returns.ABSOLUTE:
        ret_series = series - shifted_series
    else:
        raise MqValueError('Unknown returns type (use simple / logarithmic / absolute)')

    return ret_series


@plot_function
def prices(series: pd.Series, initial: int = 1, type: Returns = Returns.SIMPLE) -> pd.Series:
    """
    Calculate price levels from returns series

    :param series: time series of returns
    :param initial: initial price level
    :param type: returns type: simple, logarithmic or absolute
    :return: date-based time series of return

    **Usage**

    Compute price levels from returns series, based on the value of *type*:

    ===========   =============================
    Type          Description
    ===========   =============================
    simple        Simple arithmetic returns
    logarithmic   Logarithmic returns
    absolute      Absolute returns
    ===========   =============================

    *Simple*

    Compute asset price series from simple returns:

    :math:`Y_t = (1 + X_{t-1}) Y_{t-1}`

    where :math:`X_t` is the asset price at time :math:`t` and :math:`Y_0 = initial`

    *Logarithmic*

    Compute asset price series from logarithmic returns:

    :math:`Y_t = e^{X_{t-1}} Y_{t-1}`

    where :math:`X_t` is the asset price at time :math:`t` and :math:`Y_0 = initial`

    *Absolute*

    Compute asset price series from absolute returns:

    :math:`Y_t = X_{t-1} + Y_{t-1}`

    where :math:`X_t` is the asset price at time :math:`t` and :math:`Y_0 = initial`

    **Examples**

    Generate price series and take compute returns

    >>> series = generate_series(100)
    >>> returns = prices(returns(series))

    **See also**

    :func:`returns` :func:`product` :func:`exp`
    """

    if series.size < 1:
        return series

    if type == Returns.SIMPLE:
        return product(1 + series) * initial
    elif type == Returns.LOGARITHMIC:
        return product(series.apply(math.exp)) * initial
    elif type == Returns.ABSOLUTE:
        return sum_(series) + initial
    else:
        raise MqValueError('Unknown returns type (use simple / Logarithmic / absolute)')


@plot_function
def index(x: pd.Series, initial: int = 1) -> pd.Series:
    """
    Geometric series normalization

    :param x: time series
    :param initial: initial value
    :return: normalized time series

    **Usage**

    Divides every value in x by the initial value of x:

    :math:`Y_t = initial * X_t / X_0`

    where :math:`X_0` is the first value in the series

    **Examples**

    Normalize series to 1:

    >>> series = generate_series(100)
    >>> returns = index(series)

    **See also**

    :func:`returns`

    """
    i = x.first_valid_index()
    if not x[i]:
        raise MqValueError('Divide by zero error. Ensure that the first value of series passed to index(...) '
                           'is non-zero')
    return pd.Series(dtype=float) if i is None else initial * x / x[i]


@plot_function
def change(x: pd.Series) -> pd.Series:
    """
    Arithmetic series normalization

    :param x: time series
    :return: normalized time series

    **Usage**

    Compute difference of every value from the initial value of x:

    :math:`Y_t = X_t - X_0`

    where :math:`X_0` is the first value in the series

    **Examples**

    Change in level from initial value:

    >>> series = generate_series(100)
    >>> returns = change(series)

    **See also**

    :func:`index`

    """
    return x - x[0]


def _get_annualization_factor(x):
    prev_idx = x.index[0]
    distances = []

    for idx, value in x.iloc[1:].items():
        d = (idx - prev_idx).days
        if d == 0:
            raise MqValueError('multiple data points on same date')
        distances.append(d)
        prev_idx = idx

    average_distance = np.average(distances)
    if average_distance < 2.1:
        factor = AnnualizationFactor.DAILY
    elif 6 <= average_distance < 8:
        factor = AnnualizationFactor.WEEKLY
    elif 14 <= average_distance < 17:
        factor = AnnualizationFactor.SEMI_MONTHLY
    elif 25 <= average_distance < 35:
        factor = AnnualizationFactor.MONTHLY
    elif 85 <= average_distance < 97:
        factor = AnnualizationFactor.QUARTERLY
    elif 360 <= average_distance < 386:
        factor = AnnualizationFactor.ANNUALLY
    else:
        raise MqValueError('Cannot infer annualization factor, average distance: ' + str(average_distance))
    return factor


@plot_function
def annualize(x: pd.Series) -> pd.Series:
    """
    Annualize series based on sample observation frequency

    :param x: time series of prices
    :return: date-based time series of annualized values

    **Usage**

    Based on number of days between observations, will determine an annualization factor and then adjust values
    accordingly. Useful for annualizing daily or monthly returns

    :math:`Y_t = X_t * \\sqrt{F}`

    Annualization factors as follows, based on period implied by observations:

    =========   =============================
    Period      Annualization Factor (F)
    =========   =============================
    Daily       :math:`252`
    Weekly      :math:`52`
    Bi-Weekly   :math:`26`
    Monthly     :math:`12`
    Quarterly   :math:`4`
    Annually    :math:`1`
    =========   =============================

    **Examples**

    Annualize daily returns series:

    >>> prices = generate_series(100)
    >>> ann = annualize(returns(prices))

    **See also**

    :func:`returns`
    """

    factor: int = _get_annualization_factor(x)
    return x * math.sqrt(factor)


@plot_function
def volatility(x: pd.Series, w: Union[Window, int, str] = Window(None, 0),
               returns_type: Returns = Returns.SIMPLE) -> pd.Series:
    """
    Realized volatility of price series

    :param x: time series of prices
    :param w: Window or int: size of window and ramp up to use. e.g. Window(22, 10) where 22 is the window size
              and 10 the ramp up value.  If w is a string, it should be a relative date like '1m', '1d', etc.
              Window size defaults to length of series.
    :param returns_type: returns type: simple, logarithmic or absolute
    :return: date-based time series of return

    **Usage**

    Calculate rolling annualized realized volatility of a price series over a given window. Annual volatility of 20% is
    returned as 20.0:

    :math:`Y_t = \\sqrt{\\frac{1}{N-1} \\sum_{i=t-w+1}^t (R_t - \\overline{R_t})^2} * \\sqrt{252} * 100`

    where N is the number of observations in each rolling window :math:`w`, :math:`R_t` is the return on time
    :math:`t` based on *returns_type*

    ===========   =======================================================
    Type          Description
    ===========   =======================================================
    simple        Simple geometric change in asset prices:
                  :math:`R_t = \\frac{X_t}{X_{t-1}} - 1`
                  where :math:`X_t` is the asset price at time :math:`t`
    logarithmic   Natural logarithm of asset price changes:
                  :math:`R_t = log(X_t) - log(X_{t-1})`
                  where :math:`X_t` is the asset price at time :math:`t`
    absolute      Absolute change in asset prices:
                  :math:`Y_t = X_t - X_{t-obs}`
                  where :math:`X_t` is the asset price at time :math:`t`
    ===========   =======================================================

    and :math:`\\overline{R_t}` is the mean value over the same window:

    :math:`\\overline{R_t} = \\frac{\\sum_{i=t-w+1}^{t} R_t}{N}`

    If window is not provided, computes realized volatility over the full series

    **Examples**

    Compute rolling :math:`1` month (:math:`22` business day) annualized volatility of price series

    >>> series = generate_series(100)
    >>> vol_series = volatility(series, 22)
    >>> vol_series = volatility(series, Window(22, 30))

    **See also**

    :func:`std` :func:`annualize` :func:`returns`

    """
    w = normalize_window(x, w)

    if x.size < 1:
        return x

    return apply_ramp(annualize(std(returns(x, type=returns_type), Window(w.w, 0))).mul(100), w)


@plot_function
def correlation(x: pd.Series, y: pd.Series,
                w: Union[Window, int, str] = Window(None, 0), type_: SeriesType = SeriesType.PRICES,
                returns_type: Returns = Returns.SIMPLE) -> pd.Series:
    """
    Rolling correlation of two price series

    :param x: price series
    :param y: price series
    :param w: Window, int, or str: size of window and ramp up to use. e.g. Window(22, 10) where 22 is the window size
              and 10 the ramp up value. If w is a string, it should be a relative date like '1m', '1d', etc.
              Window size defaults to length of series.
    :param type_: type of both input series: prices or returns
    :param returns_type: Method to calculate returns when type_ is PRICES, simple, logarithmic or absolute. You can
                        also pass two methods to calculate returns for x and y respectively.
    :return: date-based time series of correlation

    **Usage**

    Calculate rolling `realized correlation <https://en.wikipedia.org/wiki/Correlation_and_dependence>`_,
    :math:`\\rho_t` of two price series over a given window:

    :math:`\\rho_t =
    \\frac{\\sum_{i=t-w+1}^t (R_t - \\overline{R_t})(Y_t - \\overline{S_t})}{(N-1)\\sigma R_t\\sigma S_t}`

    where N is the number of observations in each rolling window, :math:`w`, and :math:`R_t` and :math:`S_t` are the
    simple returns for each series on time :math:`t`

    If prices are provided then returns are calculated based on the returns_type e.g. for simple returns:

    :math:`R_t = \\frac{X_t}{X_{t-1}} - 1` and :math:`S_t = \\frac{Y_t}{Y_{t-1}} - 1`

    If returns are provided:

    :math:`R_t = X_t` and :math:`S_t = Y_t`

    :math:`\\overline{R_t}`, :math:`\\overline{S_t}` are the mean values, and :math:`\\sigma R_{t}` and
    :math:`\\sigma S_{t}` are the sample standard deviations, of  series
    :math:`R_t` and :math:`S_t` over the same window

    If window is not provided, computes realized correlation over the full series

    **Examples**

    Compute rolling :math:`1` month (:math:`22` business day) correlation of price series

    >>> series1 = generate_series(100)
    >>> series2 = generate_series(100)
    >>> corr = correlation(series1, series2, 22)

    **See also**

    :func:`std` :func:`returns`

    """
    w = normalize_window(x, w)

    if x.size < 1:
        return x

    given_prices = type_ == SeriesType.PRICES
    if given_prices:
        if isinstance(returns_type, (tuple, list)):
            if len(returns_type) != 2:
                raise MqValueError('Expected a list of length 2 for "returns_type"')
            if not all(isinstance(r, Returns) for r in returns_type):
                raise MqTypeError('Expected a list of Returns for "returns_type"')
            returns_type_x, returns_type_y = returns_type
        else:
            returns_type_x = returns_type_y = returns_type
        ret_1 = returns(x, type=returns_type_x)
        ret_2 = returns(y, type=returns_type_y)
    else:
        ret_1 = x
        ret_2 = y

    clean_ret1 = ret_1.dropna()
    clean_ret2 = ret_2.dropna()

    if isinstance(w.w, pd.DateOffset):
        if isinstance(clean_ret1.index, pd.DatetimeIndex):
            values = [clean_ret1.loc[(clean_ret1.index > idx - w.w) & (clean_ret1.index <= idx)].corr(clean_ret2)
                      for idx in clean_ret1.index]
        else:
            values = [clean_ret1.loc[(clean_ret1.index > (idx - w.w).date()) &
                                     (clean_ret1.index <= idx)].corr(clean_ret2)
                      for idx in clean_ret1.index]
        corr = pd.Series(values, index=clean_ret1.index)
    else:
        corr = clean_ret1.rolling(w.w, 0).corr(clean_ret2)

    return apply_ramp(interpolate(corr, x, Interpolate.NAN), w)


@plot_function
def beta(x: pd.Series, b: pd.Series, w: Union[Window, int, str] = Window(None, 0), prices: bool = True) -> pd.Series:
    """
    Rolling beta of price series and benchmark

    :param x: time series of prices
    :param b: time series of benchmark prices
    :param w: Window, int, or str: size of window and ramp up to use. e.g. Window(22, 10) where 22 is the window size
              and 10 the ramp up value.  If w is a string, it should be a relative date like '1m', '1d', etc.
              Window size defaults to length of series.
    :param prices: True if input series are prices, False if they are returns
    :return: date-based time series of beta

    **Usage**

    Calculate rolling `beta <https://en.wikipedia.org/wiki/Beta_(finance)>`_,
    :math:`\\beta_t` of a series to a benchmark over a given window:

    :math:`R_t = \\alpha_t + \\beta S_t + \\epsilon_t`

    Calculated as:

    :math:`\\beta_t = \\frac{\\sum_{i=t-w+1}^t Cov(R_t, S_t)}{Var(S_t)}`

    where N is the number of observations in each rolling window, :math:`w`, and :math:`R_t` and :math:`S_t` are the
    simple returns for each series on time :math:`t`:

    :math:`R_t = \\frac{X_t}{X_{t-1}} - 1` and :math:`S_t = \\frac{b_t}{b_{t-1}} - 1`

    If prices = False, assumes returns are provided:

    :math:`R_t = X_t` and :math:`S_t = b_t`

    :math:`Cov(R_t, S_t)` and :math:`Var(S_t)` are the covariance and variance of the series
    :math:`R_t` and :math:`S_t` over the same window

    If window is not provided, computes beta over the full series

    **Examples**

    Compute rolling :math:`1` month (:math:`22` business day) beta of two price series

    >>> series = generate_series(100)
    >>> benchmark = generate_series(100)
    >>> b = beta(series, benchmark, 22)

    **See also**

    :func:`var` :func:`cov` :func:`correlation` :func:`returns`
    """
    if not isinstance(prices, bool):
        raise MqTypeError('expected a boolean value for "prices"')

    w = normalize_window(x, w)

    ret_series = returns(x) if prices else x
    ret_benchmark = returns(b) if prices else b

    if isinstance(w.w, pd.DateOffset):
        series_index = ret_series.index.intersection(ret_benchmark.index)
        size = len(series_index)
        ret_series = ret_series.loc[series_index]
        benchmark_series = ret_benchmark.loc[series_index]

        ret_values = np.array(ret_series.values, dtype=np.double)
        benchmark_values = np.array(benchmark_series.values, dtype=np.double)

        cov_results = np.empty(size, dtype=np.double)
        var_results = np.empty(size, dtype=np.double)

        offset = w.w
        start = 0
        for i in range(1, size):
            min_index_value = (series_index[i] - offset).date()
            for idx in range(start, i + 1):
                if series_index[idx] > min_index_value:
                    start = idx
                    break

            sub_benchmark_values = benchmark_values[start:i + 1]
            var_results[i] = np.var(sub_benchmark_values, ddof=1)
            cov_results[i] = np.cov(ret_values[start:i + 1], sub_benchmark_values, ddof=1)[0][1]

        result = pd.Series(cov_results / var_results, index=series_index, dtype=np.double)
    else:
        cov = ret_series.rolling(w.w, 0).cov(ret_benchmark)
        result = cov / ret_benchmark.rolling(w.w, 0).var()

    # do not compute initial values as they may be extreme when sample size is small
    result[0:3] = np.nan
    return apply_ramp(interpolate(result, x, Interpolate.NAN), w)


@plot_function
def max_drawdown(x: pd.Series, w: Union[Window, int, str] = Window(None, 0)) -> pd.Series:
    """
    Compute the maximum peak to trough drawdown over a rolling window as a ratio.

    i.e. if the max drawdown for a period is 20%, this function will return -0.2.

    :param x: time series
    :param w: Window, int, or str: size of window and ramp up to use. e.g. Window(22, 10) where 22 is the window size
              and 10 the ramp up value.  If w is a string, it should be a relative date like '1m', '1d', etc.
              Window size defaults to length of series.
    :return: time series of rolling maximum drawdown

    **Examples**

    Compute the maximum peak to trough `drawdown <https://en.wikipedia.org/wiki/Drawdown_(economics)>`_

    >>> series = generate_series(100)
    >>> max_drawdown(series)

    **See also**

    :func:`returns`

    """
    w = normalize_window(x, w)
    if isinstance(w.w, pd.DateOffset):
        if pd.api.types.is_datetime64_dtype(x.index):
            scores = pd.Series([x[idx] / x.loc[(x.index > (idx - w.w)) & (x.index <= idx)].max() - 1
                                for idx in x.index], index=x.index)
            result = pd.Series([scores.loc[(scores.index > (idx - w.w)) & (scores.index <= idx)].min()
                                for idx in scores.index], index=scores.index)
        else:
            raise TypeError('Please pass in list of dates as index')
    else:
        rolling_max = x.rolling(w.w, 0).max()
        result = (x / rolling_max - 1).rolling(w.w, 0).min()
    return apply_ramp(result, w)



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import logging
import traceback
from contextlib import ContextDecorator
from enum import Enum
from typing import Tuple, Optional, Sequence, Mapping, Union

import pandas as pd
from opentelemetry import trace, context
from opentelemetry.context import Context
from opentelemetry.propagate import extract, inject, set_global_textmap
from opentelemetry.propagators.textmap import TextMapPropagator
from opentelemetry.sdk.trace import TracerProvider, SynchronousMultiSpanProcessor, ReadableSpan, Span, Event
from opentelemetry.sdk.trace.export import SimpleSpanProcessor, SpanExporter
from opentelemetry.trace import Tracer as OtelTracer, SpanContext, INVALID_SPAN
from opentelemetry.trace import format_trace_id, format_span_id

from gs_quant.errors import MqWrappedError

_logger = logging.getLogger(__name__)


class Tags(Enum):
    HTTP_METHOD = 'http.method'
    HTTP_URL = 'http.url'
    HTTP_STATUS_CODE = 'http.status_code'
    CONTENT_LENGTH = 'content.length'


class SpanConsumer(SpanExporter):
    _instance = None

    @staticmethod
    def get_instance():
        if SpanConsumer._instance is None:
            SpanConsumer._instance = SpanConsumer()
        return SpanConsumer._instance

    @staticmethod
    def get_spans() -> Sequence['TracingSpan']:
        return SpanConsumer.get_instance()._collected_spans

    @staticmethod
    def reset():
        SpanConsumer.get_instance()._collected_spans = []

    @staticmethod
    def manually_record(spans: Sequence['TracingSpan']):
        SpanConsumer.get_instance()._collected_spans.extend(spans)

    def __init__(self):
        self._collected_spans = []

    def export(self, spans: Sequence[ReadableSpan]) -> None:
        self._collected_spans.extend(TracingSpan(span) for span in spans)


class TracingContext:
    def __init__(self, ctx: SpanContext):
        self._context = ctx


class TracingEvent:

    def __init__(self, ot_event: Event):
        self._event = ot_event

    @property
    def name(self) -> str:
        return self._event.name

    @property
    def timestamp(self) -> int:
        """
        Timestamp in ns
        """
        return self._event.timestamp

    @property
    def timestamp_sec(self) -> float:
        """
        Timestamp in seconds
        """
        return self._event.timestamp / 1e9

    @property
    def attributes(self) -> Mapping[str, any]:
        return self._event.attributes


class TracingScope:

    def __init__(self, token, span: Optional[Span], finish_on_close: bool = True):
        self._token = token
        self._span = TracingSpan(span) if span else NonRecordingTracingSpan(INVALID_SPAN)
        self._finish_on_close = finish_on_close

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            Tracer.record_exception(exc_val, self._span, exc_tb)
        self.close()

    def close(self):
        if self._token:
            context.detach(self._token)
            if self._finish_on_close:
                return self._span.end()

    @property
    def span(self) -> 'TracingSpan':
        return self._span


class TracingSpan:
    def __init__(self, span: Span, endpoint: Optional[str] = None):
        self._span = span
        self._endpoint = endpoint

    def unwrap(self):
        return self._span

    @property
    def context(self):
        return self

    def end(self):
        self._span.end()

    def is_recording(self):
        return self._span.is_recording()

    @property
    def operation_name(self) -> str:
        return self._span.name

    def transportable(self, endpoint_override: Optional[str] = None) -> 'TransportableSpan':
        return TransportableSpan(self, endpoint_override or self._endpoint)

    @property
    def endpoint(self) -> str:
        return self._endpoint

    @endpoint.setter
    def endpoint(self, endpoint):
        self._endpoint = endpoint

    @property
    def trace_id(self) -> str:
        return format_trace_id(self._span.get_span_context().trace_id)

    def is_error(self) -> bool:
        return self._span.attributes.get('error', False)

    @property
    def start_time(self) -> int:
        """
        Start time in ns
        """
        return self._span.start_time

    @property
    def end_time(self) -> Optional[int]:
        """
        End time in ns
        """
        return self._span.end_time

    @property
    def duration(self) -> float:
        """
        Duration of the span in milliseconds, or None if the span is not finished.
        """
        unwrapped = self._span
        return (unwrapped.end_time - unwrapped.start_time) / 1e6 if unwrapped.end_time else None

    @property
    def span_id(self) -> str:
        return format_span_id(self._span.get_span_context().span_id)

    @property
    def parent_id(self) -> Optional[str]:
        parent = self._span.parent
        return format_span_id(parent.span_id) if parent else None

    @property
    def tags(self) -> Mapping[str, any]:
        return self._span.attributes

    @property
    def events(self) -> Sequence[TracingEvent]:
        return tuple(TracingEvent(event) for event in self._span.events)

    def set_tag(self, key: Union[Enum, str], value: Union[bool, str, bytes, int, float, dt.date]) -> 'TracingSpan':
        if value is None:
            return self
        if isinstance(value, dt.date):
            value = value.isoformat()
        elif isinstance(value, Enum):
            value = value.value
        if isinstance(key, Enum):
            key = key.value
        self._span.set_attribute(key, value)
        return self

    def add_event(self, name: str, attributes: Optional[Mapping[str, any]] = None,
                  timestamp: Optional[float] = None) -> 'TracingSpan':
        converted_timestamp = int(timestamp * 1e9) if timestamp else None
        self._span.add_event(name, attributes, converted_timestamp)
        return self

    def log_kv(self, key_values: Mapping[str, any], timestamp=None) -> 'TracingSpan':
        converted_timestamp = int(timestamp * 1e9) if timestamp else None
        event_name = "log" if key_values is None or "event" not in key_values else key_values["event"]
        self._span.add_event(event_name, key_values, converted_timestamp)
        return self


class NonRecordingTracingSpan(TracingSpan):
    def __init__(self, span: Span):
        super().__init__(span)
        self._span = span

    def end(self):
        pass

    @property
    def operation_name(self) -> str:
        return "Non-Recording Span"

    @property
    def parent_id(self) -> Optional[str]:
        return None

    @property
    def tags(self) -> Mapping[str, any]:
        return dict()

    @property
    def start_time(self) -> int:
        return 0

    @property
    def end_time(self) -> Optional[int]:
        return None

    @property
    def duration(self) -> float:
        return 0


NOOP_TRACING_SCOPE = TracingScope(None, None)


class TransportableSpan(TracingSpan):
    """
    A transportable span is a representation of a finished TracingSpan that can be pickled.
    """

    def __init__(self, span: TracingSpan, endpoint: Optional[str] = None):
        super().__init__(None, endpoint or span.endpoint)
        self._operation_name = span.operation_name
        self._trace_id = span.trace_id
        self._span_id = span.span_id
        self._parent_id = span.parent_id
        self._tags = dict(span.tags)
        self._start_time = span.start_time
        self._end_time = span.end_time
        self._events = tuple(TransportableTracingEvent(event) for event in span.events)

    def transportable(self, endpoint_override: Optional[str] = None) -> 'TransportableSpan':
        if not endpoint_override or endpoint_override == self._endpoint:
            return self
        else:
            return TransportableSpan(self, endpoint_override or self._endpoint)

    def end(self):
        pass

    def is_recording(self):
        return self._end_time is None

    @property
    def operation_name(self) -> str:
        return self._operation_name

    @property
    def trace_id(self) -> str:
        return self._trace_id

    def is_error(self) -> bool:
        return self._tags.get('error', False)

    @property
    def start_time(self) -> int:
        """
        Start time in ns
        """
        return self._start_time

    @property
    def end_time(self) -> Optional[int]:
        """
        End time in ns
        """
        return self._end_time

    @property
    def duration(self) -> float:
        """
        Duration of the span in milliseconds, or None if the span is not finished.
        """
        return (self._end_time - self._start_time) / 1e6 if self._end_time else None

    @property
    def span_id(self) -> str:
        return self._span_id

    @property
    def parent_id(self) -> Optional[str]:
        return self._parent_id

    @property
    def tags(self) -> Mapping[str, any]:
        return self._tags

    @property
    def events(self) -> Sequence[TracingEvent]:
        return self._events

    def set_tag(self, key: Union[Enum, str], value: Union[bool, str, bytes, int, float, dt.date]) -> 'TracingSpan':
        return self

    def add_event(self, name: str, attributes: Optional[Mapping[str, any]] = None,
                  timestamp: Optional[float] = None) -> 'TracingSpan':
        return self

    def log_kv(self, key_values: Mapping[str, any], timestamp=None) -> 'TracingSpan':
        return self


class TransportableTracingEvent(TracingEvent):

    def __init__(self, event: TracingEvent):
        super().__init__(None)
        self._name = event.name
        self._timestamp = event.timestamp
        self._attributes = dict(event.attributes)

    @property
    def name(self) -> str:
        return self._name

    @property
    def timestamp(self) -> int:
        """
        Timestamp in ns
        """
        return self._timestamp

    @property
    def timestamp_sec(self) -> float:
        """
        Timestamp in seconds
        """
        return self._timestamp / 1e9

    @property
    def attributes(self) -> Mapping[str, any]:
        return self._attributes


class TracerFactory:
    __tracer_instance = None

    def get(self) -> OtelTracer:
        if TracerFactory.__tracer_instance is None:
            # Define which OpenTelemetry Tracer provider implementation to use.
            span_processor = SynchronousMultiSpanProcessor()
            span_processor.add_span_processor(SimpleSpanProcessor(SpanConsumer.get_instance()))
            trace.set_tracer_provider(TracerProvider(active_span_processor=span_processor))

            # Create an OpenTelemetry Tracer.
            otel_tracer = trace.get_tracer(__name__)
            TracerFactory.__tracer_instance = otel_tracer
        return TracerFactory.__tracer_instance


class Tracer(ContextDecorator):
    __factory = TracerFactory()

    def __init__(self, label: str = 'Execution', print_on_exit: bool = False, threshold: int = None,
                 wrap_exceptions=False, parent_span: Optional[Union[TracingSpan, TracingContext]] = None):
        self.__print_on_exit = print_on_exit
        self.__label = label
        self.__threshold = threshold
        self.wrap_exceptions = wrap_exceptions
        self._parent_span = parent_span if isinstance(parent_span, TracingSpan) else None
        self._parent_ctx = parent_span if isinstance(parent_span, TracingContext) else None

    def __enter__(self):
        if self._parent_span:
            self.__parent_scope = Tracer.activate_span(self._parent_span)
        self.__scope = Tracer.start_active_span(self.__label, child_of=self._parent_ctx)
        return self.__scope

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_value:
            self.record_exception(exc_value, self.__scope.span, exc_tb)
        self.__scope.close()
        if self._parent_span:
            self.__parent_scope.close()
        if self.wrap_exceptions and exc_type is not None and not exc_type == MqWrappedError:
            raise MqWrappedError(f'Unable to calculate: {self.__label}') from exc_value

    @staticmethod
    def get_instance() -> OtelTracer:
        return Tracer.__factory.get()

    @staticmethod
    def set_factory(factory: TracerFactory):
        Tracer.__factory = factory

    @staticmethod
    def active_span():
        current_span = trace.get_current_span()
        if current_span is None or not current_span.is_recording():
            return NonRecordingTracingSpan(INVALID_SPAN)
        else:
            return TracingSpan(current_span)

    @staticmethod
    def set_propagator_format(propagator_format: TextMapPropagator):
        set_global_textmap(propagator_format)

    @staticmethod
    def inject(carrier):
        span = trace.get_current_span()
        if span is not None and span.is_recording():
            try:
                inject(carrier)
            except Exception:
                _logger.error("Error injecting trace context", exc_info=True)

    @staticmethod
    def extract(carrier):
        try:
            return TracingContext(extract(carrier))
        except Exception:
            _logger.error("Error extracting trace context", exc_info=True)

    @staticmethod
    def activate_span(span: TracingSpan = None, finish_on_close: bool = False) -> Optional[TracingScope]:
        """
        Activates the current span
        :return: The current span
        """
        if span is None or not span.is_recording():
            return NOOP_TRACING_SCOPE
        ctx = trace.set_span_in_context(span.unwrap())
        token = context.attach(ctx)
        return TracingScope(token, span.unwrap(), finish_on_close=finish_on_close)

    @staticmethod
    def start_active_span(operation_name: str, child_of: Optional[TracingContext] = None,
                          ignore_active_span: bool = False, finish_on_close: bool = True) -> TracingScope:
        ctx = Context() if ignore_active_span else child_of._context if child_of else None
        span = Tracer.get_instance().start_span(operation_name, context=ctx)
        # Set as the implicit current context
        # Creates a Context object with parent set as current span
        ctx = trace.set_span_in_context(span)
        token = context.attach(ctx)
        return TracingScope(token, span, finish_on_close)

    @staticmethod
    def record_exception(e, span: TracingSpan = None, exc_tb=None):
        span = span or TracingSpan(trace.get_current_span())
        if span is not None:
            try:
                span.set_tag('error', True)
                span.log_kv({
                    'event': 'error',
                    'message': str(e),
                    'error.object': str(e),
                    'error.kind': type(e).__name__,
                    'stack': Tracer.__format_traceback(type(e), e, exc_tb),
                })
            except Exception:
                pass

    @staticmethod
    def __format_traceback(exc_type, exc_value, exc_tb=None):
        if exc_value is None:
            return ''
        try:
            return ''.join(traceback.format_exception(exc_type, exc_value, exc_tb, limit=10))
        except Exception:
            return ''

    @staticmethod
    def reset():
        SpanConsumer.reset()

    @staticmethod
    def get_spans() -> Sequence[TracingSpan]:
        return SpanConsumer.get_spans()

    @staticmethod
    def plot(reset=False, show=True):
        try:
            import plotly.express as px
        except ImportError:
            _logger.warning('Package "plotly" required to visualise the trace, printing instead')
            Tracer.print(reset)
            return
        color_list = ('#2b76f7', '#5a8efb', '#79aaff', '#89bbff', '#cdddff')
        error_color = 'rgb(244, 127, 114)'
        ordered_spans, _ = Tracer.gather_data(False)
        span_df = pd.DataFrame.from_records([(
            f'#{i}',
            f'{s.operation_name} {int(s.duration):,.0f}ms',
            dt.datetime.fromtimestamp(s.start_time / 1e9),
            dt.datetime.fromtimestamp(s.end_time / 1e9),
            '\n '.join([f'{k}={v}' for k, v in s.tags.items()]) if s.tags else '',
        ) for i, (depth, s) in enumerate(ordered_spans)], columns=['id', 'operation', 'start', 'end', 'tags'])
        color_map = {f'#{i}': error_color if 'error' in s.tags else color_list[depth % len(color_list)] for
                     i, (depth, s) in enumerate(ordered_spans)}
        fig = px.timeline(
            data_frame=span_df,
            width=1000,
            height=40 + 30 * len(span_df),
            x_start="start",
            x_end="end",
            y="id",
            hover_data='tags',
            text='operation',
            color='id',
            color_discrete_map=color_map
        )
        fig.update_layout(
            showlegend=False,
            yaxis_visible=False,
            yaxis_showticklabels=False,
            paper_bgcolor='rgba(0,0,0,0)',
            plot_bgcolor='rgba(0,0,0,0)',
            margin={'l': 0, 't': 0, 'b': 0, 'r': 0}
        )
        if reset:
            Tracer.reset()
        if show:
            fig.show()
        else:
            return fig

    @staticmethod
    def gather_data(as_string: bool = True, root_id: Optional[str] = None, trace_id: Optional[str] = None):
        spans = Tracer.get_spans()
        spans_by_parent = {}

        for span in reversed(spans):
            if trace_id and span.trace_id != trace_id:
                continue
            spans_by_parent.setdefault(span.parent_id, []).append(span)

        def _build_tree(parent_span, depth):
            if as_string:
                name = f'{"* " * depth}{parent_span.operation_name}'
                elapsed = (parent_span.end_time - parent_span.start_time) / 1000000
                error = " [Error]" if parent_span.is_error() else ""
                lines.append(f'{name:<50}{elapsed:>8.1f} ms{error}')
            else:
                lines.append((depth, parent_span))
            for child_span in reversed(spans_by_parent.get(parent_span.span_id, [])):
                _build_tree(child_span, depth + 1)

        total = 0
        lines = []
        # By default, we look for the span with no parent, but this might not always be what we want
        for span in reversed(spans_by_parent.get(root_id, [])):
            _build_tree(span, 0)
            total += (span.end_time - span.start_time) / 1000000

        if as_string:
            tracing_str = '\n'.join(lines)
            return tracing_str, total
        else:
            return lines, total

    @staticmethod
    def print(reset=True, root_id=None, trace_id=None):
        tracing_str, total = Tracer.gather_data(root_id=root_id, trace_id=trace_id)
        str_id = trace_id or root_id or ""
        _logger.warning(f'Tracing Info: {str_id}\n{tracing_str}\n{"-" * 61}\nTOTAL:{total:>52.1f} ms')
        if reset:
            Tracer.reset()
        return tracing_str, total


def parse_tracing_line_args(line: str) -> Tuple[Optional[str], bool]:
    stripped = tuple(s for s in line.split(' ') if s != '')
    if len(stripped) > 0 and stripped[0] in ('chart', 'plot', 'graph'):
        return tuple(stripped[1:]) if len(stripped[1:]) else None, True
    return stripped if len(stripped) else None, False


try:
    # Attempt to import/register some jupyter magic
    import gs_quant_internal.tracing.jupyter  # noqa
except ImportError:
    try:
        from IPython.core.magic import register_cell_magic
        from IPython import get_ipython

        @register_cell_magic("trace")
        def trace_ipython_cell(line, cell):
            """Wraps the execution of a cell in a tracer call and prints"""
            span_name, show_chart = parse_tracing_line_args(line)
            if cell is None:
                return line
            with Tracer(label=span_name):
                res = get_ipython().run_cell(cell)
                if res.error_in_exec:
                    Tracer.record_exception(res.error_in_exec)
            if show_chart:
                Tracer.plot(True)
            else:
                Tracer.print(True)
            return None
    except Exception:
        pass



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt

from gs_quant.api.gs.esg import GsEsgApi, ESGCard, ESGMeasure
from gs_quant.session import Environment, GsSession


def test_get_risk_models(mocker):
    mock_response = {
        'pricingDate': '2021-04-08',
        'summary': {
            'gross': {
                'gPercentile': 97.57,
                'gRegionalPercentile': 88.94,
                'esPercentile': 89.52,
                'esDisclosurePercentage': 83.33,
                'esMomentumPercentile': 28.15
            },
            'long': {
                'gPercentile': 97.57,
                'gRegionalPercentile': 88.94,
                'esPercentile': 89.52,
                'esDisclosurePercentage': 83.33,
                'esMomentumPercentile': 28.15},
            'short': {
                'gPercentile': None,
                'gRegionalPercentile': None,
                'esPercentile': None,
                'esDisclosurePercentage': None,
                'esMomentumPercentile': None
            },
            'benchmark': {
                'gPercentile': 76.30923829640506,
                'gRegionalPercentile': 54.3718544735162,
                'esPercentile': 51.677217204614415,
                'esDisclosurePercentage': 54.89969169789053,
                'esMomentumPercentile': 48.186284955203256
            }
        },
        'bottomTenRanked': [
            {
                'measure': 'gPercentile',
                'results': [
                    {
                        'assetId': 'MA4B66MW5E27UAHKG34',
                        'name': 'The Goldman Sachs Group, Inc.',
                        'value': 97.57
                    }
                ]
            }, {
                'measure': 'esDisclosurePercentage',
                'results': [
                    {
                        'assetId': 'MA4B66MW5E27UAHKG34',
                        'name': 'The Goldman Sachs Group, Inc.',
                        'value': 83.33
                    }
                ]
            }
        ],
        'noESGData': {
            'gross': {
                'weight': 0.0,
                'assets': []
            },
            'long': {
                'weight': 0.0,
                'assets': []
            },
            'short': {
                'weight': 0.0,
                'assets': []
            },
            'benchmark': {
                'weight': 2.5402270308391683,
                'assets': [
                    {
                        'assetId': 'MA4B66MW5E27UALNB96',
                        'name': 'News Corp - Class B',
                        'weight': 0.026140889953506573
                    }, {
                        'assetId': 'MA4B66MW5E27UANLXX3',
                        'name': 'Under Armour Inc-Class C',
                        'weight': 0.02019587639265163
                    }, {
                        'assetId': 'MA4B66MW5E27UAGYZ49',
                        'name': 'Alphabet Inc-CL C',
                        'weight': 2.4141476728491242
                    }, {
                        'assetId': 'MA4B66MW5E27UAE4MGB',
                        'name': 'Discovery Inc-C',
                        'weight': 0.03970496421509619
                    }, {
                        'assetId': 'MA4B66MW5E27UAGEHL4',
                        'name': 'Fox Corp - Class A',
                        'weight': 0.040037627428790246
                    }
                ]
            }
        }
    }

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsEsgApi.get_esg(entity_id='MPM9Y3434RDFVKA3',
                                benchmark_id='MA4B66MW5E27U8P32SB',
                                pricing_date=dt.date(2021, 4, 8),
                                measures=[ESGMeasure.G_PERCENTILE, ESGMeasure.ES_DISCLOSURE_PERCENTAGE],
                                cards=[ESGCard.SUMMARY, ESGCard.BOTTOM_TEN_RANKED])
    GsSession.current._get.assert_called_with(
        '/esg/MPM9Y3434RDFVKA3?&date=2021-04-08&benchmark=MA4B66MW5E27U8P32SB&measure=gPercentile&'
        'measure=esDisclosurePercentage&card=summary&card=bottomTenRanked')
    assert response == mock_response



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from enum import Enum
from gs_quant.target.secmaster import SecMasterAssetType
from gs_quant.common import AssetType, AssetClass
from typing import Union, Iterable, Dict, Optional
from gs_quant.json_encoder import JSONEncoder
from gs_quant.session import GsSession
import json
import datetime as dt

SECURITIES_FEDERATED = '/markets/securities/federated'


class FederatedIdentifiers(Enum):
    IDENTIFIER = 'identifier'
    ID = 'id'
    ASSET_ID = 'assetId'
    GSID = 'gsid'
    TICKER = 'ticker'
    BBID = 'bbid'
    BCID = 'bcid'
    RIC = 'ric'
    RCIC = 'rcic'
    CUSIP = 'cusip'
    CINS = 'cins'
    SEDOL = 'sedol'
    ISIN = 'isin'
    PRIMEID = 'primeId'


class GsSecurityMasterFederatedApi:

    @classmethod
    def get_a_security(cls, id: str, effective_date: dt.date = None) -> Optional[dict]:
        """
        Get security or asset for given security id or asset id.
        @param id: secmaster id(e.g. 'GSPD111E123') or asset id(e.g. 'MANYS1FCCWWV45P7')
        @param effective_date: As of date for query
        @return: dict
        """

        if not id.startswith("GS") and not id.startswith("MA"):
            raise ValueError(f"Invalid id: {id}. Security id starts with 'GS' and Asset id starts with 'MA'")

        params = {
        }
        if effective_date is not None:
            params["effectiveDate"] = effective_date
        payload = json.loads(json.dumps(params, cls=JSONEncoder))

        return GsSession.current._get(f'{SECURITIES_FEDERATED}/{id}', payload=payload)

    @classmethod
    def get_security_identifiers(cls, id: str) -> Optional[dict]:
        """
        Get identifiers history for given security id or asset id.
        @param id: secmaster id(e.g. 'GSPD111E123') or asset id(e.g. 'MANYS1FCCWWV45P7')
        @return: dict
        """

        if not id.startswith("GS") and not id.startswith("MA"):
            raise ValueError(f"Invalid id: {id}. Security id starts with 'GS' and Asset id starts with 'MA'")

        return GsSession.current._get(f'{SECURITIES_FEDERATED}/{id}/identifiers')

    @classmethod
    def get_many_securities(cls,
                            type_: Union[SecMasterAssetType, AssetType] = None,
                            effective_date: dt.date = None,
                            limit: int = 50,
                            is_primary=None,
                            offset_key: str = None,
                            **query_params: Dict[FederatedIdentifiers, Union[str, Iterable[str]]]) -> Optional[dict]:
        """
        Get reference data for a single page. Use returned offsetKey to fetch next page.
        @param is_primary:
        @param type_: security or asset type
        @param effective_date: As of date for query
        @param limit: integer of individual page
        @param offset_key: string, an offset indicating where the page ends
        @return: dict
        """

        return cls.__query_securities(type_=type_,
                                      effective_date=effective_date,
                                      limit=limit,
                                      flatten=False,
                                      is_primary=is_primary,
                                      offset_key=offset_key,
                                      **query_params)

    @classmethod
    def get_securities_data(cls,
                            type_: Union[SecMasterAssetType, AssetType] = None,
                            effective_date: dt.date = None,
                            limit: int = 50,
                            is_primary=None,
                            offset_key: str = None,
                            **query_params: Dict[FederatedIdentifiers, Union[str, Iterable[str]]]) -> Optional[dict]:
        """
        Get flattened reference data for a single page. Use returned offsetKey to fetch next page.
        @param is_primary:
        @param type_: security or asset type
        @param effective_date: As of date for query
        @param limit: integer of individual page
        @param offset_key: string, an offset indicating where the page ends
        @return: dict
        """

        return cls.__query_securities(type_=type_,
                                      effective_date=effective_date,
                                      limit=limit,
                                      flatten=True,
                                      is_primary=is_primary,
                                      offset_key=offset_key,
                                      **query_params)

    @classmethod
    def search_many_securities(cls,
                               q: str = None,
                               limit: int = 10,
                               offset_key: str = None,
                               asset_class: AssetClass = None,
                               type_: Union[SecMasterAssetType, AssetType] = None,
                               is_primary: bool = None) -> Optional[dict]:
        """
        Search reference data by a query string. It does a full text search among names, identifiers, company.
        @param q: query string
        @param limit: number of returned matches
        @param offset_key: string, an offset indicating where the page ends
        @param asset_class: filter restricting the class of results
        @param type_: filter restricting the type of results
        @param is_primary: filter restricting the matches to primary listings
        @return: dict
        """

        return cls.__search_securities(q=q,
                                       limit=limit,
                                       offset_key=offset_key,
                                       flatten=False,
                                       asset_class=asset_class,
                                       type_=type_,
                                       is_primary=is_primary)

    @classmethod
    def search_securities_data(cls,
                               q: str = None,
                               limit: int = 10,
                               offset_key: str = None,
                               asset_class: AssetClass = None,
                               type_: Union[SecMasterAssetType, AssetType] = None,
                               is_primary: bool = None) -> Optional[dict]:
        """
        Search flattened reference data by a query string. It does a full text search among names, identifiers, company.
        @param q: query string
        @param limit: number of returned matches
        @param offset_key: string, an offset indicating where the page ends
        @param asset_class: filter restricting the class of results
        @param type_: filter restricting the type of results
        @param is_primary: filter restricting the matches to primary listings
        @return: dict
        """

        return cls.__search_securities(q=q,
                                       limit=limit,
                                       offset_key=offset_key,
                                       flatten=True,
                                       asset_class=asset_class,
                                       type_=type_,
                                       is_primary=is_primary)

    @classmethod
    def __query_securities(cls,
                           type_: Union[SecMasterAssetType, AssetType] = None,
                           effective_date: dt.date = None,
                           limit: int = 50,
                           flatten=False,
                           is_primary=None,
                           offset_key: str = None,
                           **query_params: Dict[FederatedIdentifiers, Union[str, Iterable[str]]]
                           ) -> Optional[dict]:
        if (query_params is None or len(query_params) == 0) and type_ is None:
            raise ValueError("Neither '_type' nor 'query_params' are provided")

        params = {
            "limit": limit
        }
        cls.__prepare_params(params, effective_date, offset_key, is_primary, type_, None)
        params = {**params, **query_params}
        payload = json.loads(json.dumps(params, cls=JSONEncoder))
        if flatten:
            return GsSession.current._get(f'{SECURITIES_FEDERATED}/data', payload=payload)
        return GsSession.current._get(f'{SECURITIES_FEDERATED}', payload=payload)

    @classmethod
    def __search_securities(cls,
                            q: str = None,
                            limit: int = 10,
                            offset_key: str = None,
                            flatten=False,
                            asset_class: AssetClass = None,
                            type_: Union[SecMasterAssetType, AssetType] = None,
                            is_primary: bool = None) -> Optional[dict]:
        if (q is None):
            raise ValueError("No search query provided")

        params = {
            "q": q,
            "limit": limit
        }
        cls.__prepare_params(params, None, offset_key, is_primary, type_, asset_class)
        payload = json.loads(json.dumps(params, cls=JSONEncoder))
        if flatten:
            return GsSession.current._get(f'{SECURITIES_FEDERATED}/search/data', payload=payload)
        return GsSession.current._get(f'{SECURITIES_FEDERATED}/search', payload=payload)

    @classmethod
    def __prepare_params(cls, params, effective_date, offset_key, is_primary, type_, asset_class):
        if effective_date is not None:
            params["effectiveDate"] = effective_date
        if offset_key is not None:
            params["offsetKey"] = offset_key
        if is_primary is not None:
            params["isPrimary"] = is_primary
        if type_ is not None:
            params["type"] = type_.value
        if asset_class is not None:
            params["assetClass"] = asset_class.value



"""
Copyright 2023 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import json
import math
from enum import Enum
from functools import partial
from itertools import groupby
from typing import Union, Iterable, Dict, Optional

import tqdm

from gs_quant.data.utilities import SecmasterXrefFormatter
from gs_quant.json_encoder import JSONEncoder
from gs_quant.session import GsSession
from gs_quant.target.secmaster import SecMasterAssetType

DEFAULT_SCROLL_PAGE_SIZE = 500


class SecMasterIdentifiers(Enum):
    CUSIP = 'cusip'
    TICKER = 'ticker'
    ISIN = 'isin'
    GSID = 'gsid'
    BBG = 'bbg'
    BBID = 'bbid'
    BCID = 'bcid'
    RIC = 'ric'
    RCIC = 'rcic'
    ID = 'id'
    ASSET_ID = 'assetId'
    CUSIP8 = 'cusip8'
    SEDOL = 'sedol'
    CINS = 'cins'
    PRIMEID = 'primeId'
    FACTSET_REGIONAL_ID = 'factSetRegionalId'
    TOKEN_ID = 'tokenId'
    COMPOSITE_FIGI = 'compositeFigi'
    BARRA_ID = 'barraId'
    AXIOMA_ID = 'axiomaId'
    FIGI = 'figi'


def __extend_enum(base_enum, new_values):
    members = {item.name: item.value for item in base_enum}
    members.update(new_values)
    return Enum('CapitalStructureIdentifiers', members)


# FIXME Create an enum in service def for this and reference once module_generator get fixed.
CapitalStructureIdentifiers = __extend_enum(SecMasterIdentifiers, {"ISSUER_ID": "issuerId"})


class ExchangeId(Enum):
    RIC_SUFFIX_CODE = "ricSuffixCode"
    RIC_EXCHANGE_CODE = "ricExchangeCode"
    DATASCOPE_IPC_CODE = "datascopeIpcCode"
    BBG_EXCHANGE_CODE = "bbgExchangeCode"
    TAQ_EXCHANGE_CODE = "taqExchangeCode"
    IVERSON_EXCHANGE_CODE = "iversonExchangeCode"
    INDEX_CHANGE_EXCHANGE_CODE = "indexChangeExchangeCode"
    DOW_JONES_EXCHANGE_CODE = "dowJonesExchangeCode"
    STOXX_EXCHANGE_CODE = "stoxxExchangeCode"
    ML_ETF_EXCHANGE_CODE = "mlEtfExchangeCode"
    FTSE_EXCHANGE_CODE = "ftseExchangeCode"
    DJGI_EXCHANGE_CODE = "djgiExchangeCode"
    SECDB_EXCHANGE_CODE = "secdbExchangeCode"
    DADD_EXCHANGE_CODE = "daddExchangeCode"
    MIC = "mic"
    OPERATING_MIC = "operatingMic"
    GS_EXCHANGE_ID = "gsExchangeId"
    COUNTRY = "country"
    EXCHANGE_NAME = "name"


class GsSecurityMasterApi:

    @classmethod
    def get_security(cls, id_value: str,
                     id_type: SecMasterIdentifiers,
                     effective_date: dt.date = None):
        """
        Get flatten asset reference data

        @param id_value: identifier value
        @param id_type: identifier type
        @param effective_date: As of date for query
        @return: dict or None
        """
        args = {id_type.value: id_value}
        results = cls.get_many_securities(effective_date=effective_date, flatten=False, **args)
        if results is not None:
            return results["results"][0]
        return results

    @classmethod
    def get_many_securities(cls, type_: SecMasterAssetType = None,
                            effective_date: dt.date = None,
                            limit: int = 10, flatten=False,
                            is_primary=None,
                            offset_key: str = None,
                            **query_params: Dict[SecMasterIdentifiers, Union[str, Iterable[str]]]) \
            -> Optional[dict]:
        """
        Get reference data for a single page of a given asset type. Use returned offsetKey to fetch next page.

        @param is_primary:
        @param flatten: flag if data should be flattened
        @param type_: asset type
        @param effective_date: As of date for query
        @param limit: integer of individual page
        @param offset_key: string, an offset indicating where the page ends.
        @return: list of dict
        """

        if (query_params is None or len(query_params) == 0) and type_ is None:
            raise ValueError("Neither '_type' nor 'query_params' are provided")

        params = {
            "limit": limit
        }

        cls.prepare_params(params, is_primary, offset_key, type_, effective_date)

        params = {**params, **query_params}
        payload = json.loads(json.dumps(params, cls=JSONEncoder))

        if flatten:
            r = GsSession.current._get('/markets/securities/data', payload=payload)
        else:
            r = GsSession.current._get('/markets/securities', payload=payload)
        if r['totalResults'] == 0:
            return None
        return r

    @classmethod
    def get_all_securities(cls, type_: SecMasterAssetType = None,
                           effective_date: dt.date = None,
                           is_primary=None,
                           flatten=False, **query_params) -> Optional[dict]:
        """
        Get all securities reference data matching the type, with respect of effective_date property.
        Function runs in batches fetching all securities.

        @param flatten: Flag, whether data should be flattened
        @param is_primary: Restrict to primary listings
        @param type_: asset type
        @param effective_date: As of date for query
        @return:" list of dict
        """
        if 'limit' in query_params:
            limit = query_params['limit']
            del query_params['limit']
        else:
            limit = DEFAULT_SCROLL_PAGE_SIZE

        response = cls.get_many_securities(type_, effective_date, limit=limit, offset_key=None,
                                           flatten=flatten, is_primary=is_primary,
                                           **query_params)
        if response is None or "offsetKey" not in response:
            return response

        if response['totalResults'] == 0:
            return None

        results = response["results"]
        offset_key = response["offsetKey"]

        fn = partial(cls.get_many_securities, type_=type_, effective_date=effective_date,
                     limit=limit, flatten=flatten,
                     **query_params)
        results.extend(cls.__fetch_all(fn, offset_key))
        response["totalResults"] = len(results)
        response["results"] = results

        return response

    @classmethod
    def get_security_data(cls, id_value: str,
                          id_type: SecMasterIdentifiers,
                          effective_date: dt.date = None) -> Optional[dict]:
        """
        Get flatten asset reference data

        @param id_value: identifier value
        @param id_type: identifier type
        @param effective_date: As of date for query
        @return: dict or None
        """
        args = {id_type.value: id_value}
        results = cls.get_many_securities(effective_date=effective_date, flatten=True, **args)
        if results is not None:
            return results["results"][0]
        return results

    @classmethod
    def get_identifiers(cls, secmaster_id: str) -> dict:
        """
        Get identifiers history for given secmaster id.

        @param secmaster_id:  secmaster id  e.g. ['GSPD111E123']
        @return: list of dict with date ranges of the identifiers.
        """
        if not secmaster_id.startswith("GS"):
            raise ValueError(f"Invalid id_value {secmaster_id}. Secmaster id starts with 'GS'")
        r = GsSession.current._get(f'/markets/securities/{secmaster_id}/identifiers')
        return r['results']

    @classmethod
    def get_many_identifiers(cls, ids: Iterable[str], limit=100, xref_format=False) -> dict:
        """
          Get identifiers for a list of secmaster ids. It runs in batches till all data is fetched.

          This method retrieves identifier information for multiple securities.
          The data can be returned in either standard format or transformed using the SecmasterXrefFormatter
          for time-based cross-reference analysis.

          Args:
              ids (Iterable[str]): An iterable collection of secmaster identifiers to query.
                                  Examples: ['GSPD111E123', 'GSPD222F456', 'GSPD333G789']

              limit (int, optional): Maximum number of identifier records to return per secmaster id.
                                    Defaults to 100. Used to control response size and prevent
                                    memory issues with large datasets.

              xref_format (bool, optional): Flag to control output format. Defaults to False.
                                          - False: Returns raw identifier data in standard format
                                          - True: Returns data transformed by SecmasterXrefFormatter,
                                                 where identifiers are grouped by overlapping date
                                                 ranges

          Returns:
              dict: A dictionary containing identifier information for the requested secmaster ids.

                    Standard format (xref_format=False):
                    {
                        'secmaster_id_1': [
                            {
                                'type': 'ISIN',
                                'value': 'US1234567890',
                                'startDate': '2020-01-01',
                                'endDate': '2023-12-31'
                            },
                            ...
                        ],
                        'secmaster_id_2': [...],
                        ...
                    }

                    Xref format (xref_format=True):
                    {
                        'secmaster_id_1': {
                            'xrefs': [
                                {
                                    'startDate': '2020-01-01',
                                    'endDate': '2023-12-31',
                                    'identifiers': [
                                        {'type': 'ISIN', 'value': 'US1234567890'},
                                        {'type': 'CUSIP', 'value': '123456789'},
                                        ...
                                    ]
                                },
                                ...
                            ]
                        },
                        'secmaster_id_2': {...},
                        ...
                    }

          Raises:
              ValueError: If ids parameter is empty or contains invalid secmaster identifiers
              ConnectionError: If unable to connect to secmaster database
              TimeoutError: If database query exceeds timeout limits

          Note:
              - The method processes requests in batches to handle large datasets efficiently
              - Infinity dates ('9999-99-99') are normalized to '9999-12-31' in xref format


          Example:
              >>> # Standard format
              >>> result = GsSecurityMasterApi.get_many_identifiers(['GSPD111E123'], limit=50, xref_format=False)
              >>> print(result['GSPD111E123'][0]['type'])  # 'ISIN'

              >>> # Xref format for time-based analysis
              >>> xref_result = GsSecurityMasterApi.get_many_identifiers(['GSPD111E123'], xref_format=True)
              >>> print(xref_result['GSPD111E123']['xrefs'][0]['startDate'])  # '2020-01-01'
          """
        if not isinstance(ids, Iterable):
            raise ValueError(f"secmaster_id must be an iterable, got {type(ids)}")
        if len(ids) == 0:
            raise ValueError("secmaster_id cannot be an empty iterable")
        ids = list(ids)

        for id_value in ids:
            if not id_value.startswith("GS"):
                raise ValueError(f"Invalid id_value {id_value}. Secmaster id starts with 'GS'")

        consolidated_results = {}
        current_offset_key = None

        while True:
            payload = {'id': ids}

            if current_offset_key is not None:
                payload['offsetKey'] = current_offset_key
            if limit is not None:
                payload['limit'] = limit

            payload = json.loads(json.dumps(payload, cls=JSONEncoder))
            response = GsSession.current._get('/markets/securities/identifiers', payload=payload)

            if 'results' in response:
                for entity_id, data in response['results'].items():
                    if entity_id not in consolidated_results:
                        consolidated_results[entity_id] = []
                    consolidated_results[entity_id].extend(data)

            current_offset_key = response.get('offsetKey')
            if current_offset_key is None:
                break

        if xref_format:
            return SecmasterXrefFormatter.convert(consolidated_results)
        return consolidated_results

    @classmethod
    def map(cls, input_type: SecMasterIdentifiers,
            ids: Iterable[str],
            output_types: Iterable[SecMasterIdentifiers] = frozenset([SecMasterIdentifiers.GSID]),
            start_date: dt.date = None,
            end_date: dt.date = None,
            effective_date: dt.date = None) -> Iterable[dict]:
        """
        Map to other identifier types, from given IDs.

        :param input_type: type of input IDs
        :param ids: security IDs
        :param output_types: types of IDs to map to
        :param start_date: first as-of date (defaults to current date)
        :param end_date: last as-of date (defaults to current date)
        :param effective_date: an exact as-of date for mapping
        :return: dict containing mappings for as-of date(s)
        """
        params = {
            input_type.value: list(ids),
            'toIdentifiers': [identifier.value for identifier in output_types],
            'compact': True
        }
        if effective_date is not None:
            if (start_date or end_date) is not None:
                raise ValueError('provide (start date / end date) or effective_date, but not both')
            params['effectiveDate'] = effective_date

        if start_date is not None:
            params['startDate'] = start_date
        if end_date is not None:
            params['endDate'] = end_date
        payload = json.loads(json.dumps(params, cls=JSONEncoder))

        r = GsSession.current._get('/markets/securities/map', payload)
        results = r['results']
        return results

    @classmethod
    def search(cls, q: str, limit: int = 10, type_: SecMasterAssetType = None, is_primary: bool = None,
               active_listing: bool = None) -> Union[Iterable[dict], None]:
        """
        Search securities by a query string. It does a full text search among names, identifiers, company
        @param q: query string
        @param limit: number of returned matches
        @param type_: filter restricting the type of results
        @param is_primary: filter restricting the matches to primary listings
        @param active_listing: filter restricting the matches to active listings
        @return:
        """
        params = {
            "q": q,
            "limit": limit
        }
        if type_ is not None:
            params["type"] = type_.value
        if is_primary is not None:
            params["isPrimary"] = is_primary
        if active_listing is not None:
            params["activeListing"] = active_listing
        payload = json.loads(json.dumps(params, cls=JSONEncoder))
        r = GsSession.current._get('/v2/markets/securities/search', payload=payload, include_version=False)
        if r['totalResults'] == 0:
            return None
        return r["results"]

    @classmethod
    def __stringify_boolean(cls, bool_value):
        return str(bool_value).lower()

    @classmethod
    def __fetch_all(cls, fetch_fn, offset_key, total_batches=None, extract_results=True):
        accumulator = []
        offset = offset_key
        progress_info = tqdm.tqdm(desc="Processing", unit=" batch") if total_batches is None else tqdm.tqdm(
            range(total_batches), desc="Processing", unit=" batch", ascii=True)
        while True:
            progress_info.update(1)
            data = fetch_fn(offset_key=offset)
            if data is not None:
                if extract_results is True:
                    accumulator.extend(data['results'])
                else:
                    accumulator.append(data)
                if 'offsetKey' not in data:
                    progress_info.close()
                    break
                offset = data["offsetKey"]
        return accumulator

    @classmethod
    def _get_corporate_actions(cls, id_value: str, id_type: SecMasterIdentifiers,
                               effective_date: dt.date, offset_key):

        params = {
            id_type.value: id_value,
        }
        if effective_date is not None:
            params['effectiveDate'] = effective_date

        if offset_key is not None:
            params["offsetKey"] = offset_key
        payload = json.loads(json.dumps(params, cls=JSONEncoder))
        r = GsSession.current._get("/markets/corpactions", payload=payload)
        return r

    @classmethod
    def get_corporate_actions(cls, id_value: str, id_type: SecMasterIdentifiers = SecMasterIdentifiers.GSID,
                              effective_date: dt.date = None) -> Iterable[dict]:
        """
        Get corporate actions from a given security.
        @param effective_date: parameter to query securities at a given date
        @param id_value: identifier value
        @param id_type: identifier type
        @return:
        """
        supported_identifiers = [SecMasterIdentifiers.GSID, SecMasterIdentifiers.ID]
        if id_type not in supported_identifiers:
            raise ValueError(
                f"Unsupported identifier {id_type} for this endpoint. Use one of this {supported_identifiers}")

        fn = partial(cls._get_corporate_actions, id_value, id_type, effective_date)
        results = cls.__fetch_all(fn, None)
        return results

    @classmethod
    def get_capital_structure(cls, id_value: Union[str, list],
                              id_type: CapitalStructureIdentifiers,
                              type_: SecMasterAssetType = None, is_primary: bool = None,
                              effective_date: dt.date = None) -> dict:
        """
        Get a capital structure of the given company by id_value of the security.
         It runs in batches till all data is  fetched
        @param is_primary: filter to select primary listings only
        @param type_:  filter to restrict data to a given type
        @param effective_date: parameter to query securities at a given date
        @param id_value: identifier value
        @param id_type: identifier type
        @return: dict
        """
        response = cls._get_capital_structure(id_value=id_value, id_type=id_type, type_=type_, is_primary=is_primary,
                                              effective_date=effective_date, offset_key=None)
        if "offsetKey" not in response:
            return response

        asset_types_total = response["assetTypesTotal"]
        batch_count = math.floor(sum(asset_types_total.values()) / 100)
        results = response["results"]
        offset_key = response["offsetKey"]
        fn = partial(cls._get_capital_structure, id_value, id_type, type_, is_primary, effective_date)

        results.extend(cls.__fetch_all(fn, offset_key, total_batches=batch_count))
        aggregated_results, total_results = cls.__capital_structure_aggregate(asset_types_total, results)
        response["results"] = aggregated_results
        response["totalResults"] = total_results
        del response["offsetKey"]

        return response

    @classmethod
    def __capital_structure_aggregate(cls, asset_types_total, results):
        group_by_issuer_id = {k: list(g) for k, g in groupby(results, lambda x: x["issuerId"])}
        aggregated_results = []
        aggregated_total_results = 0
        for issuer_id in group_by_issuer_id:
            issuer_id_data = group_by_issuer_id[issuer_id]
            issuer_id_data_instance = issuer_id_data[0]
            consolidated_types_obj = {key: [] for key in asset_types_total}
            for obj in group_by_issuer_id[issuer_id]:
                aggregated_total_results += sum(len(e) for e in obj["types"].values())
                [consolidated_types_obj.get(asset_type).extend(obj["types"].get(asset_type))
                 for asset_type in obj["types"]]
                issuer_id_data_instance["types"] = consolidated_types_obj
            aggregated_results.append(issuer_id_data_instance)
        return aggregated_results, aggregated_total_results

    @classmethod
    def _get_capital_structure(cls, id_value: Union[str, list],
                               id_type: Union[CapitalStructureIdentifiers, SecMasterIdentifiers],
                               type_, is_primary, effective_date, offset_key: Union[str, None]):
        params = {
            id_type.value: id_value
        }
        cls.prepare_params(params, is_primary, offset_key, type_, effective_date)
        payload = json.loads(json.dumps(params, cls=JSONEncoder))
        r = GsSession.current._get("/markets/capitalstructure", payload=payload)
        return r

    @classmethod
    def prepare_params(cls, params, is_primary, offset_key, type_, effective_date=None):
        if type_ is not None:
            params["type"] = type_.value
        if is_primary is not None:
            params["isPrimary"] = is_primary
        if offset_key is not None:
            params["offsetKey"] = offset_key
        if effective_date is not None:
            params["effectiveDate"] = effective_date

    @classmethod
    def _get_deltas(cls, start_time: dt.datetime = None, end_time: dt.datetime = None, raw: bool = None,
                    scope: list = None, limit: int = None, offset_key: str = None) -> \
            Iterable[dict]:

        params = {}
        if raw is not None:
            params["raw"] = GsSecurityMasterApi.__stringify_boolean(raw)
        if start_time is not None:
            params["startTime"] = start_time
        if end_time is not None:
            params["endTime"] = end_time
        if scope is not None:
            params["scope"] = scope
        if limit is not None:
            params["limit"] = limit
        if offset_key is not None:
            params["offsetKey"] = offset_key

        payload = json.loads(json.dumps(params, cls=JSONEncoder))
        r = GsSession.current._get("/markets/securities/identifiers/updates-feed", payload=payload)
        return r

    @classmethod
    def get_deltas(cls, start_time: dt.datetime = None, end_time: dt.datetime = None, raw: bool = None,
                   scope: list = None, limit: int = None, offset_key: str = None, scroll_all_pages: bool = True) -> \
            Union[dict, Iterable[dict]]:
        """
        Get all identifier changes between two time stamps
        @param scroll_all_pages:
        @param start_time: start time
        @param end_time: end time
        @param limit: page size of returned matches
        @param scope: narrow down the search to a specific set of events
        @param offset_key: offset key to fetch next page
        @param raw: flag, if true (default) aggregates data to more readable form, if false shows unprocessed results.
        @return: list of dict
        """
        if scroll_all_pages:
            fn = partial(cls._get_deltas, start_time, end_time, raw, scope, limit)
            results = cls.__fetch_all(fn, offset_key, extract_results=False)
            latest_update_time = max(result['lastUpdateTime'] for result in results)
            res = [item for result in results for item in result["results"]]
            request_id = results[0]["requestId"] if results else None
            return {"results": res, "lastUpdateTime": latest_update_time, "requestId": request_id}
        else:
            results = cls._get_deltas(start_time, end_time, raw, scope, limit, offset_key)
        return results

    @classmethod
    def get_exchanges(cls, effective_date: dt.date = None,
                      **query_params: Dict[str, Union[str, Iterable[str]]]):
        """
        Returns reference data for exchanges - e.g. MICs, exchange codes, name, listing country.

        @param effective_date: As of date for query
        @param query_params: one of allowed params:
         'ricSuffixCode', 'ricExchangeCode', 'datascopeIpcCode', 'bbgExchangeCode', 'taqExchangeCode',
        'iversonExchangeCode', 'indexChangeExchangeCode', 'dowJonesExchangeCode', 'stoxxExchangeCode',
        'mlEtfExchangeCode', 'ftseExchangeCode', 'djgiExchangeCode', 'secdbExchangeCode', 'daddExchangeCode', 'mic',
        'operatingMic', 'gsExchangeId', 'country', 'name'
        @return:
        """
        results = []
        fn = partial(cls._get_exchanges, effective_date, DEFAULT_SCROLL_PAGE_SIZE, query_params)
        results.extend(cls.__fetch_all(fn, offset_key=None))
        response = dict()
        response["totalResults"] = len(results)
        response["results"] = results

        return response

    @classmethod
    def _get_exchanges(cls, effective_date: dt.date = None, limit: int = 10,
                       query_params=None,
                       offset_key: Union[str, None] = None):

        if query_params is None:
            query_params = dict()
        allowed_keys = list(ExchangeId._value2member_map_.keys())
        for qp in query_params.keys():
            if qp not in allowed_keys:
                raise ValueError(f" Parameter '{qp}' is not supported. Allowed parameters:  {allowed_keys}")
        params = {
            "limit": limit
        }
        if effective_date is not None:
            params['effectiveDate'] = effective_date

        params = {**params, **query_params}
        if offset_key is not None:
            params["offsetKey"] = offset_key
        payload = json.loads(json.dumps(params, cls=JSONEncoder))
        r = GsSession.current._get('/markets/exchanges', payload=payload)
        if r['totalResults'] == 0:
            return None
        return r

    @classmethod
    def get_exchange_identifiers_history(cls, gs_exchange_id: str) -> Iterable[dict]:
        """
        Get identifiers history for given exchange id.

        @param gs_exchange_id:  exchange_id id
        @return: list of dict with date ranges of the identifiers.
        """
        r = GsSession.current._get(f'/markets/exchanges/{gs_exchange_id}/identifiers')
        return r['results']

    @classmethod
    def _prepare_string_or_list_param(cls, value: Union[str, list], param_name: str):
        if isinstance(value, str):
            return [value]  # Wrap single string in a list
        elif isinstance(value, list) and all(isinstance(item, str) for item in value):
            return value  # Use the list directly
        else:
            raise ValueError(f"{param_name} must be a string or a list of strings")

    @classmethod
    def _prepare_underlyers_params(cls, params, id_value, offset_key=None, type_=None, effective_date=None,
                                   country_code=None, currency=None, include_inactive=None):
        if id_value is not None:
            params["id"] = cls._prepare_string_or_list_param(id_value, "id_value")
        else:
            raise ValueError("id_value must be defined")

        if type_ is not None:
            if isinstance(type_, list):
                if all(isinstance(t, SecMasterAssetType) for t in type_):
                    params["type"] = [t.value for t in type_]
                else:
                    raise ValueError("All elements in the type_ list must be instances of SecMasterAssetType")
            elif isinstance(type_, SecMasterAssetType):
                params["type"] = type_.value
            else:
                raise ValueError("type_ must be either a SecMasterAssetType or a list of SecMasterAssetType")

        if country_code is not None:
            params["countryCode"] = cls._prepare_string_or_list_param(country_code, "country_code")
        if currency is not None:
            params["currency"] = cls._prepare_string_or_list_param(currency, "currency")
        if offset_key is not None:
            params["offsetKey"] = offset_key
        if effective_date is not None:
            params["effectiveDate"] = effective_date
        if include_inactive is not None:
            params["includeInactive"] = include_inactive

    @classmethod
    def _get_securities_by_underlyers(cls,
                                      id_value: Union[str, list],
                                      type_: Union[SecMasterAssetType, list] = None,
                                      effective_date: dt.date = None,
                                      limit: int = 100,
                                      country_code: Union[str, list] = None,
                                      currency: Union[str, list] = None,
                                      include_inactive: bool = False,
                                      offset_key: str = None) \
            -> Optional[dict]:
        params = {
            "limit": limit
        }

        cls._prepare_underlyers_params(
            params=params,
            id_value=id_value,
            offset_key=offset_key,
            type_=type_,
            effective_date=effective_date,
            country_code=country_code,
            currency=currency,
            include_inactive=include_inactive
        )

        payload = json.loads(json.dumps(params, cls=JSONEncoder))

        r = GsSession.current._get('/markets/securities/underlyers', payload=payload)
        if r['totalResults'] == 0:
            return None
        return r

    @classmethod
    def get_securities_by_underlyers(cls,
                                     id_value: Union[str, list],
                                     type_: Union[SecMasterAssetType, list] = None,
                                     effective_date: dt.date = None,
                                     limit: int = None,
                                     offset_key: str = None,
                                     country_code: Union[str, list] = None,
                                     currency: Union[str, list] = None,
                                     include_inactive: bool = False,
                                     scroll_all_pages: bool = False) \
            -> Optional[dict]:
        """
        Retrieve reference data for listed derivatives based on their underlyers.

        This method retrieves securities linked to specific underlyers, with optional filters for asset type, effective
        date, country code and currency. The results can be paginated, and the method supports fetching either
        a single page or all pages of results.

        Args:
            id_value (Union[str, list]): Identifier(s) of the underlyers. Can be a single string or a list of strings.
                                          Example: 'GSPD100E0' or ['GSPD100E0'].
            type_ (Union[SecMasterAssetType, list], optional): Asset type(s) to filter the results. Can be a single
                                                               `SecMasterAssetType` or a list of `SecMasterAssetType`.
                                                               Defaults to None.
            effective_date (dt.date, optional): Effective date for the query. Defaults to None.
            limit (int, optional): Maximum number of results to return per page. Defaults to None.
            offset_key (str, optional): String indicating where the page ends, used for pagination. Defaults to None.
            country_code (Union[str, list], optional): Country code(s) to filter the results. Can be a single string
                                                       or a list of strings. Defaults to None.
            currency (Union[str, list], optional): Currency code(s) to filter the results. Can be a single string
                                                   or a list of strings. Defaults to None.
            include_inactive (bool, optional): Whether to include inactive listed derivatives in the results.
                                               Defaults to False.
            scroll_all_pages (bool, optional): Whether to fetch all pages of results. If True, retrieves all pages.
                                               Defaults to False.

        Returns:
            Optional[dict]: A dictionary containing the results for the requested page(s), or None if no results are
                            found.
                            Example:
                            {
                                "totalResults": 10,
                                "offsetKey": "ABCD=",
                                "results": [...],
                                "asOfTime": "2025-01-11T11:22:33.740Z",
                                "requestId": "abc123"
                            }

        Raises:
            ValueError: If `id_value` is not provided or is invalid.
            ValueError: If `type_` is not a valid `SecMasterAssetType` or a list of valid `SecMasterAssetType`.
            ValueError: If `country_code` or `currency` is not a string or a list of strings.
            ValueError: If `effective_date` is not in the format 'YYYY-MM-DD' when provided as a string.
            TypeError: If `effective_date` is not of type `datetime.date` or `str`.

        Note:
            - This method fetches a single page of results by default. To fetch all pages, set `scroll_all_pages=True`.
            - The `effective_date` parameter is used to query securities as of a specific date.
            - The `offset_key` parameter is used for pagination to fetch subsequent pages of results.

        Example:
            >>> result = GsSecurityMasterApi.get_securities_by_underlyers(
            ...     id_value="GSPD100E0",
            ...     type_=SecMasterAssetType.Equity_Option,
            ...     effective_date="2023-10-01",
            ...     country_code="US",
            ...     currency="USD",
            ...     include_inactive=False,
            ...     scroll_all_pages=True
            ... )
            >>> print(result["results"])
        """

        supported_types = [SecMasterAssetType.Equity_Option, SecMasterAssetType.Future,
                           SecMasterAssetType.Future_Option]
        if type_ is not None:
            if isinstance(type_, SecMasterAssetType):
                if type_ not in supported_types:
                    raise ValueError(f"Unsupported type {type_}. Supported types are {supported_types}")
            elif isinstance(type_, list):
                for t in type_:
                    if t not in supported_types:
                        raise ValueError(f"Unsupported type {t}. Supported types are {supported_types}")

        if scroll_all_pages:
            if limit is None:
                limit = 500
            fn = partial(cls._get_securities_by_underlyers,
                         id_value=id_value,
                         type_=type_,
                         effective_date=effective_date,
                         limit=limit,
                         country_code=country_code,
                         currency=currency,
                         include_inactive=include_inactive)
            results = cls.__fetch_all(fn, offset_key, extract_results=False)
            as_of_time = max(result['asOfTime'] for result in results)
            res = [item for result in results for item in result["results"]]
            request_id = results[0]["requestId"] if results else None
            total_results = len(res)
            return {"totalResults": total_results, "results": res, "asOfTime": as_of_time, "requestId": request_id}

        else:
            return cls._get_securities_by_underlyers(
                id_value=id_value,
                type_=type_,
                effective_date=effective_date,
                limit=limit,
                offset_key=offset_key,
                country_code=country_code,
                currency=currency,
                include_inactive=include_inactive
            )



"""
Copyright 2024 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import json

from freezegun import freeze_time

from gs_quant.common import PricingLocation
from gs_quant.json_encoder import JSONEncoder
from gs_quant.markets import CloseMarket, PricingContext


def test_close_market_dict():
    mkt = CloseMarket(date=dt.date(2024, 4, 11))
    assert json.dumps(mkt.to_dict(), cls=JSONEncoder) == json.dumps(mkt.market.to_dict(), cls=JSONEncoder)
    mkt = CloseMarket(date=dt.date(2024, 4, 11), location='LDN')
    assert json.dumps(mkt.to_dict(), cls=JSONEncoder) == json.dumps(mkt.market.to_dict(), cls=JSONEncoder)


@freeze_time("2025-03-20 16:00:00", tz_offset=4)
def test_close_market_roll():
    # 8PM in LDN, 4PM in NYC
    CloseMarket.roll_hr_and_min = (17, 30)
    test_date = dt.date(2025, 3, 20)
    with PricingContext(test_date, market_data_location="NYC"):
        assert PricingContext.current.market == CloseMarket(test_date - dt.timedelta(days=1), PricingLocation.NYC)
    CloseMarket.roll_hr_and_min = (12, 00)
    with PricingContext(test_date, market_data_location="NYC"):
        assert PricingContext.current.market == CloseMarket(test_date, PricingLocation.NYC)


@freeze_time("2025-03-20 22:00:00", tz_offset=4)
def test_close_market_roll_diff_days():
    # 2AM in LDN, 10PM yesterday in NYC
    CloseMarket.roll_hr_and_min = (23, 30)
    test_date = dt.date(2025, 3, 21)
    with PricingContext(test_date, market_data_location="NYC"):
        # requested a date in the future for NYC - roll back to yesterday and do not apply the roll logic
        assert PricingContext.current.market == CloseMarket(test_date - dt.timedelta(days=1), PricingLocation.NYC)
    test_date = dt.date(2025, 3, 20)
    with PricingContext(test_date, market_data_location="NYC"):
        # requested today NYC, we are not past the roll time so roll back to 2025-03-19
        assert PricingContext.current.market == CloseMarket(test_date - dt.timedelta(days=1), PricingLocation.NYC)



"""
Copyright 2020 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
from unittest.mock import Mock

import pandas as pd
import pytest
from pandas.testing import assert_series_equal
from testfixtures import Replacer

import gs_quant.timeseries.measures_factset as tm
from gs_quant.api.gs.data import MarketDataResponseFrame
from gs_quant.data import DataContext
from gs_quant.errors import MqValueError
from gs_quant.markets.index import Index
from gs_quant.timeseries.measures_factset import (EstimateItem, EstimateStatistic, EstimateBasis, FiscalPeriod,
                                                  FundamentalMetric, FundamentalFormat, FundamentalBasis, RatingType)

_index = [pd.Timestamp('2021-03-30')]
_test_datasets = ('TEST_DATASET',)


def mock_fe_estimate_af(_cls, bbid, start, end, feItem):
    d = {
        'date': [
            dt.date(2022, 1, 8),
            dt.date(2022, 1, 12),
            dt.date(2022, 1, 13)
        ],
        'isin': ['US0378331005'] * 3,
        'feItem': ['EPS'] * 3,
        'fePerRel': [1.0] * 3,
        'adjDate': [dt.date(2020, 8, 31)] * 3,
        'currency': ['USD'] * 3,
        'consEndDate': [dt.date(2022, 1, 11)] * 3,
        'feItemDesc': ['Earnings Per Share'] * 3,
        'feFpEnd': [dt.date(2022, 9, 30)] * 3,
        'feDownAf': [10.0, 3.0, 3.0],
        'feHighAf': [6.353628] * 3,
        'feLowAf': [5.2] * 3,
        'feMeanAf': [5.749689, 5.748439, 5.756439],
        'feMedianAf': [5.745] * 3,
        'feNumEstAf': [40] * 3,
        'feStdDevAf': [0.259250, 0.258130, 0.258687],
        'feUpAf': [15.0, 5.0, 6.0],
        'fsymId': ['MH33D6-R'] * 3,
        'bbid': ['AAPL UW'] * 3,

    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FE_BASIC_CONH_AF_GLOBAL'

    return df


def mock_fe_estimate_qf(_cls, bbid, start, end, feItem):
    d = {
        'date': [
            dt.date(2022, 1, 8),
            dt.date(2022, 1, 12),
            dt.date(2022, 1, 13)
        ],
        'isin': ['US0378331005'] * 3,
        'feItem': ['EPS'] * 3,
        'fePerRel': [1.0] * 3,
        'adjDate': [dt.date(2020, 8, 31)] * 3,
        'currency': ['USD'] * 3,
        'consEndDate': [dt.date(2022, 1, 11)] * 3,
        'feItemDesc': ['Earnings Per Share'] * 3,
        'feFpEnd': [dt.date(2022, 9, 30)] * 3,
        'feDownQf': [10.0, 3.0, 3.0],
        'feHighQf': [6.353628] * 3,
        'feLowQf': [5.2] * 3,
        'feMeanQf': [5.749689, 5.748439, 5.756439],
        'feMedianQf': [5.745] * 3,
        'feNumEstQf': [40] * 3,
        'feStdDevQf': [0.259250, 0.258130, 0.258687],
        'feUpQf': [15.0, 5.0, 6.0],
        'fsymId': ['MH33D6-R'] * 3,
        'bbid': ['AAPL UW'] * 3,

    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FE_BASIC_CONH_QF_GLOBAL'
    return df


def mock_fe_estimate_saf(_cls, bbid, start, end, feItem):
    d = {
        'date': [
            dt.date(2022, 1, 8),
            dt.date(2022, 1, 12),
            dt.date(2022, 1, 13)
        ],
        'isin': ['US0378331005'] * 3,
        'feItem': ['EPS'] * 3,
        'fePerRel': [1.0] * 3,
        'adjDate': [dt.date(2020, 8, 31)] * 3,
        'currency': ['USD'] * 3,
        'consEndDate': [dt.date(2022, 1, 11)] * 3,
        'feItemDesc': ['Earnings Per Share'] * 3,
        'feFpEnd': [dt.date(2022, 9, 30)] * 3,
        'feDownSaf': [10.0, 3.0, 3.0],
        'feHighSaf': [6.353628] * 3,
        'feLowSaf': [5.2] * 3,
        'feMeanSaf': [5.749689, 5.748439, 5.756439],
        'feMedianSaf': [5.745] * 3,
        'feNumEstSaf': [40] * 3,
        'feStdDevSaf': [0.259250, 0.258130, 0.258687],
        'feUpSaf': [15.0, 5.0, 6.0],
        'fsymId': ['MH33D6-R'] * 3,
        'bbid': ['AAPL UW'] * 3,

    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FE_BASIC_CONH_SAF_GLOBAL'
    return df


def mock_fe_estimate_ntm(_cls, bbid, start, end, feItem):
    d = {
        'date': [
            dt.date(2022, 1, 8),
            dt.date(2022, 1, 9),
            dt.date(2022, 1, 10)
        ],
        'isin': ['US0378331005'] * 3,
        'feItem': ['EPS'] * 3,
        'feHighNtm': [6.353628] * 3,
        'feLowNtm': [5.2] * 3,
        'feMeanNtm': [5.749689, 5.748439, 5.756439],
        'feMedianNtm': [5.745] * 3,
        'feHighStm': [6.353628] * 3,
        'feLowStm': [5.2] * 3,
        'feMeanStm': [5.749689, 5.748439, 5.756439],
        'feMedianStm': [5.745] * 3,
        'fsymId': ['MH33D6-R'] * 3,
        'bbid': ['AAPL UW'] * 3,

    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FE_NTM'
    return df


def mock_fe_estimate_lt(_cls, bbid, start, end, feItem):
    d = {
        'date': [
            dt.date(2022, 1, 8),
            dt.date(2022, 1, 12),
            dt.date(2022, 1, 13)
        ],
        'isin': ['US0378331005'] * 3,
        'feItem': ['PRICE_TGT'] * 3,
        'adjDate': [dt.date(2020, 8, 31)] * 3,
        'consEndDate': [dt.date(2022, 1, 11)] * 3,
        'feDownLt': [10.0, 3.0, 3.0],
        'feHighLt': [6.353628] * 3,
        'feLowLt': [5.2] * 3,
        'feMeanLt': [5.749689, 5.748439, 5.756439],
        'feMedianLt': [5.745] * 3,
        'feNumEstLt': [40] * 3,
        'feStdDevLt': [0.259250, 0.258130, 0.258687],
        'feUpLt': [15.0, 5.0, 6.0],
        'fsymId': ['MH33D6-R'] * 3,
        'bbid': ['AAPL UW'] * 3,

    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FE_BASIC_CONH_LT_GLOBAL'
    return df


def mock_fe_actual(_cls, bbid, start, end, feItem):
    d = {
        'date': [
            dt.date(2020, 1, 8),
            dt.date(2021, 1, 12),
            dt.date(2022, 1, 13)
        ],
        'isin': ['US0378331005'] * 3,
        'feItem': ['EPS'] * 3,
        'adjDate': [dt.date(2020, 8, 31)] * 3,
        'currency': ['USD'] * 3,
        'feFpEnd': [dt.date(2022, 9, 30),
                    dt.date(2021, 9, 30),
                    dt.date(2020, 9, 30)
                    ],
        'feValue': [15.0, 5.0, 6.0],
        'fsymId': ['MH33D6-R'] * 3,
        'bbid': ['AAPL UW'] * 3,

    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FE_BASIC_ACT_AF_GLOBAL'

    return df


def mock_fe_estimate_empty(_cls, bbid, start, end, feItem):
    df = MarketDataResponseFrame()
    df.dataset_id = 'FE_BASIC_CONH_SAF_GLOBAL'
    return df


def mock_factset_fundamentals_empty(_cls, bbid, start, end):
    df = MarketDataResponseFrame()
    df.dataset_id = 'FF_BASIC_R_AF_GLOBAL'
    return df


def mock_factset_fundamentals_basic(_cls, bbid, start, end):
    d = {
        'date': [pd.Timestamp('2024-09-30 00:00:00')],
        'isin': ['US0378331005'],
        'adjDate': [pd.Timestamp('2020-08-31 00:00:00')],
        'currency': ['USD'],
        'ffEpsBasic': [6.109],
        'fsymId': ['MH33D6-R'],
        'bbid': ['AAPL UW']
    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FF_BASIC_AF_GLOBAL'
    return df


def mock_factset_fundamentals_basic_derived(_cls, bbid, start, end):
    d = {
        'date': [pd.Timestamp('2024-09-30 00:00:00')],
        'isin': ['US0378331005'],
        'adjDate': [pd.Timestamp('2020-08-31 00:00:00')],
        'currency': ['USD'],
        'ffEbitdaOper': [6.109],
        'fsymId': ['MH33D6-R'],
        'bbid': ['AAPL UW']
    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FF_BASIC_DER_AF_GLOBAL'
    return df


def mock_factset_fundamentals_basic_restated(_cls, bbid, start, end):
    d = {
        'date': [pd.Timestamp('2024-09-30 00:00:00')],
        'isin': ['US0378331005'],
        'adjDate': [pd.Timestamp('2020-08-31 00:00:00')],
        'currency': ['USD'],
        'ffEpsBasic': [6.109],
        'fsymId': ['MH33D6-R'],
        'bbid': ['AAPL UW']
    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FF_BASIC_R_AF_GLOBAL'
    return df


def mock_factset_ratings(_cls, bbid, start, end):
    d = {
        'date': [pd.Timestamp('2024-12-31 00:00:00')] * 3,
        'isin': ['US0378331005'] * 3,
        'feItem': ['REC'] * 3,
        'adjDate': [dt.date(2024, 12, 31)] * 3,
        'consEndDate': [dt.date(2024, 12, 31)] * 3,
        'feItemDesc': ['Recommendation'] * 3,
        'feFpEnd': [dt.date(2024, 12, 31)] * 3,
        'feBuy': [21.0, 3.0, 3.0],
        'feHold': [14.0, 3.0, 3.0],
        'feNoRec': [1.0, 3.0, 3.0],
        'feOver': [10.0, 3.0, 3.0],
        'feSell': [4.0, 3.0, 3.0],
        'feUnder': [1.0, 3.0, 3.0],
        'feTotal': [10.0, 3.0, 3.0],
        'feMark': [1.57] * 3,
        'fsymId': ['MH33D6-R'] * 3,
        'bbid': ['AAPL UW'] * 3,
    }
    df = MarketDataResponseFrame(data=d)
    df.dataset_id = 'FE_BASIC_CONH_REC_GLOBAL'
    return df


mock_asset = Index(
    id_='MA000',
    asset_class='Equity',
    name='test',
    currency='USD',
    exchange='NASD',
)


def test_factset_estimates():
    replace = Replacer()
    bbid_mock = replace('gs_quant.markets.securities.Asset.get_identifier', Mock())
    bbid_mock.return_value = 'AAPL UW'
    with DataContext(start=dt.date(2022, 1, 1),
                     end=dt.date(2025, 1, 31)):
        replace('gs_quant.data.Dataset.get_data', mock_fe_estimate_af)

        actual = tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                      statistic=EstimateStatistic.MEDIAN,
                                      report_basis=EstimateBasis.ANN,
                                      period=FiscalPeriod(2022),
                                      )
        assert_series_equal(pd.Series([5.745] * 4,
                                      index=pd.DatetimeIndex(['2022-01-08', '2022-01-09', '2022-01-10', '2022-01-11'],
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=EstimateStatistic.MEDIAN.value), pd.Series(actual))
        assert actual.dataset_ids == 'FE_BASIC_CONH_AF_GLOBAL'

        replace('gs_quant.data.Dataset.get_data', mock_fe_estimate_qf)

        actual = tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                      statistic=EstimateStatistic.MEDIAN,
                                      report_basis=EstimateBasis.QTR,
                                      period=FiscalPeriod(2022, 3),
                                      )
        assert_series_equal(pd.Series([5.745] * 4,
                                      index=pd.DatetimeIndex(['2022-01-08', '2022-01-09', '2022-01-10', '2022-01-11'],
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=EstimateStatistic.MEDIAN.value), pd.Series(actual))
        assert actual.dataset_ids == 'FE_BASIC_CONH_QF_GLOBAL'

        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                 statistic=EstimateStatistic.MEDIAN,
                                 report_basis=EstimateBasis.QTR,
                                 period=FiscalPeriod(2022, 1)
                                 )

        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                 statistic=EstimateStatistic.MEDIAN,
                                 report_basis=EstimateBasis.QTR,
                                 period=FiscalPeriod(2022, -1)
                                 )

        replace('gs_quant.data.Dataset.get_data', mock_fe_estimate_saf)

        actual = tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                      statistic=EstimateStatistic.MEDIAN,
                                      report_basis=EstimateBasis.SEMI,
                                      period=FiscalPeriod(2022, 2),
                                      )
        assert_series_equal(pd.Series([5.745] * 4,
                                      index=pd.DatetimeIndex(['2022-01-08', '2022-01-09', '2022-01-10', '2022-01-11'],
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=EstimateStatistic.MEDIAN.value), pd.Series(actual))
        assert actual.dataset_ids == 'FE_BASIC_CONH_SAF_GLOBAL'

        replace('gs_quant.data.Dataset.get_data', mock_fe_estimate_ntm)

        actual = tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                      statistic=EstimateStatistic.MEDIAN,
                                      report_basis=EstimateBasis.NTM
                                      )
        assert_series_equal(pd.Series([5.745] * 3,
                                      index=pd.DatetimeIndex(['2022-01-08', '2022-01-09', '2022-01-10'],
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=EstimateStatistic.MEDIAN.value), pd.Series(actual))
        assert actual.dataset_ids == 'FE_NTM'

        actual = tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                      statistic=EstimateStatistic.MEDIAN,
                                      report_basis=EstimateBasis.STM
                                      )
        assert_series_equal(pd.Series([5.745] * 3,
                                      index=pd.DatetimeIndex(['2022-01-08', '2022-01-09', '2022-01-10'],
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=EstimateStatistic.MEDIAN.value), pd.Series(actual))
        assert actual.dataset_ids == 'FE_NTM'

        replace('gs_quant.data.Dataset.get_data', mock_fe_estimate_lt)

        actual = tm.factset_estimates(mock_asset, metric=EstimateItem.PRICE_TGT,
                                      statistic=EstimateStatistic.MEDIAN
                                      )
        assert_series_equal(pd.Series([5.745] * 4,
                                      index=pd.DatetimeIndex(['2022-01-08', '2022-01-09', '2022-01-10', '2022-01-11'],
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=EstimateStatistic.MEDIAN.value), pd.Series(actual))
        assert actual.dataset_ids == 'FE_BASIC_CONH_LT_GLOBAL'

        replace('gs_quant.data.Dataset.get_data', mock_fe_actual)

        actual = tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                      statistic=EstimateStatistic.ACTUAL,
                                      report_basis=EstimateBasis.ANN
                                      )
        assert_series_equal(pd.Series([15.0],
                                      index=pd.DatetimeIndex(['2022-09-30'], dtype='datetime64[ns]', name='date',
                                                             freq=None),
                                      name=EstimateStatistic.ACTUAL.value), pd.Series(actual))
        assert actual.dataset_ids == 'FE_BASIC_ACT_AF_GLOBAL'

        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                 statistic=EstimateStatistic.MEDIAN,
                                 report_basis=EstimateBasis.SEMI,
                                 period=FiscalPeriod(2022)
                                 )
        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                 statistic=EstimateStatistic.MEDIAN,
                                 report_basis=EstimateBasis.SEMI,
                                 period=FiscalPeriod(2022, -1)
                                 )

        replace('gs_quant.data.Dataset.get_data', None)
        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                 statistic=EstimateStatistic.MEDIAN,
                                 report_basis=EstimateBasis.SEMI,
                                 period=FiscalPeriod(2022, 1)
                                 )

        replace('gs_quant.data.Dataset.get_data', mock_fe_estimate_af)
        actual = tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                      statistic=EstimateStatistic.MEDIAN,
                                      report_basis=EstimateBasis.ANN,
                                      period=1,
                                      )
        assert_series_equal(pd.Series([5.745] * 4,
                                      index=pd.DatetimeIndex([
                                          '2022-01-08', '2022-01-09', '2022-01-10', '2022-01-11'
                                      ], dtype='datetime64[ns]', name='date', freq=None),
                                      name=EstimateStatistic.MEDIAN.value), pd.Series(actual))
        assert actual.dataset_ids == 'FE_BASIC_CONH_AF_GLOBAL'

        # Get quarterly data without specifying period
        replace('gs_quant.data.Dataset.get_data', mock_fe_estimate_qf)
        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                 statistic=EstimateStatistic.MEDIAN,
                                 report_basis=EstimateBasis.QTR,
                                 period=FiscalPeriod(2022),
                                 )

        # Invalid estimate basis
        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                 statistic=EstimateStatistic.MEDIAN,
                                 report_basis='INV',
                                 period=FiscalPeriod(2022, 1),
                                 )
        # Invalid report basis for actuals
        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                 statistic=EstimateStatistic.ACTUAL,
                                 report_basis=EstimateBasis.NTM,
                                 )
        # Invalid metric for actuals
        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.PRICE_TGT,
                                 statistic=EstimateStatistic.ACTUAL,
                                 report_basis=EstimateBasis.ANN,
                                 )

        # Get empty data response
        replace('gs_quant.data.Dataset.get_data', mock_fe_estimate_empty)
        with pytest.raises(MqValueError):
            tm.factset_estimates(mock_asset, metric=EstimateItem.EPS,
                                 statistic=EstimateStatistic.MEDIAN,
                                 report_basis=EstimateBasis.SEMI,
                                 period=FiscalPeriod(2022, 1),
                                 )

    replace.restore()


def test_factset_fundamentals():
    replace = Replacer()
    bbid_mock = replace('gs_quant.markets.securities.Asset.get_identifier', Mock())
    bbid_mock.return_value = 'AAPL UW'
    with DataContext(start=dt.date(2024, 9, 30),
                     end=dt.date(2024, 9, 30)):
        replace('gs_quant.data.Dataset.get_data', mock_factset_fundamentals_basic)

        actual = tm.factset_fundamentals(mock_asset, metric=FundamentalMetric.EPS_BASIC,
                                         report_basis=FundamentalBasis.ANN,
                                         report_format=FundamentalFormat.NON_RESTATED
                                         )
        assert_series_equal(pd.Series([6.109],
                                      index=pd.DatetimeIndex(['2024-09-30'],
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=FundamentalMetric.EPS_BASIC.value), pd.Series(actual))
        assert actual.dataset_ids == 'FF_BASIC_AF_GLOBAL'

        replace('gs_quant.data.Dataset.get_data', mock_factset_fundamentals_basic_derived)

        actual = tm.factset_fundamentals(mock_asset, metric=FundamentalMetric.EBITDA_OPER,
                                         report_basis=FundamentalBasis.ANN,
                                         report_format=FundamentalFormat.NON_RESTATED
                                         )
        assert_series_equal(pd.Series([6.109],
                                      index=pd.DatetimeIndex(['2024-09-30'],
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=FundamentalMetric.EBITDA_OPER.value), pd.Series(actual))
        assert actual.dataset_ids == 'FF_BASIC_DER_AF_GLOBAL'

        replace('gs_quant.data.Dataset.get_data', mock_factset_fundamentals_basic_restated)

        actual = tm.factset_fundamentals(mock_asset, metric=FundamentalMetric.EPS_BASIC,
                                         report_basis=FundamentalBasis.ANN,
                                         report_format=FundamentalFormat.RESTATED
                                         )
        assert_series_equal(pd.Series([6.109],
                                      index=pd.DatetimeIndex(['2024-09-30'],
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=FundamentalMetric.EPS_BASIC.value), pd.Series(actual))
        assert actual.dataset_ids == 'FF_BASIC_R_AF_GLOBAL'

        # Get empty data response
        replace('gs_quant.data.Dataset.get_data', mock_factset_fundamentals_empty)
        with pytest.raises(MqValueError):
            tm.factset_fundamentals(mock_asset, metric=FundamentalMetric.EPS_BASIC,
                                    report_basis=FundamentalBasis.ANN,
                                    report_format=FundamentalFormat.RESTATED
                                    )

    replace.restore()


def test_factset_ratings():
    replace = Replacer()
    bbid_mock = replace('gs_quant.markets.securities.Asset.get_identifier', Mock())
    bbid_mock.return_value = 'AAPL UW'
    with DataContext(start=dt.date(2024, 9, 30),
                     end=dt.date(2024, 9, 30)):
        replace('gs_quant.data.Dataset.get_data', mock_factset_ratings)

        actual = tm.factset_ratings(mock_asset, rating_type=RatingType.BUY)
        assert_series_equal(pd.Series([21.0, 3.0, 3.0],
                                      index=pd.DatetimeIndex([dt.date(2024, 12, 31)] * 3,
                                                             dtype='datetime64[ns]', name='date', freq=None),
                                      name=RatingType.BUY.value), pd.Series(actual))
        assert actual.dataset_ids == 'FE_BASIC_CONH_REC_GLOBAL'

    replace.restore()


def test_fiscal_period():
    fp = FiscalPeriod.from_dict({'y': 2022, 'p': 1})
    assert fp.y == 2022
    assert fp.p == 1

    assert fp.as_dict() == {'y': 2022, 'p': 1}


if __name__ == '__main__':
    pytest.main(args=["test_measures_factset.py"])



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import copy

from enum import Enum
from typing import Tuple, Dict, Optional, Union

from gs_quant.base import RiskMeasureParameter
from gs_quant.common import RiskMeasure, ParameterisedRiskMeasure

from gs_quant import common
from gs_quant import risk


def gsq_rm_for_name(name: str) -> Optional[RiskMeasure]:
    if name is None or name not in dir(risk):
        return None
    return getattr(risk, name)


def encode_risk_measure(rm: RiskMeasure) -> Dict:
    result = rm.as_dict(as_camel_case=True)
    if rm.parameters is not None:
        result['parameters'] = rm.parameters.as_dict(as_camel_case=True)
    return result


def encode_risk_measure_tuple(blob: Tuple[RiskMeasure, ...]) -> Tuple[Dict, ...]:
    return tuple(encode_risk_measure(rm) for rm in blob)


def _decode_param(data: dict) -> Optional[RiskMeasureParameter]:
    params = data.get('parameters', None)
    if params is not None and isinstance(params, dict) and 'parameterType' in params:
        cls_name = params['parameterType'] + 'Parameter'
        parameter_cls = getattr(common, cls_name)
        parameter = parameter_cls(**{k: v for k, v in params.items() if k != 'parameterType'})
        return parameter
    return None


def _decode_gsq_risk_measure(data: dict) -> Optional[RiskMeasure]:
    def _enum_or_str_equal(a: Optional[Union[Enum, str]], b: Optional[Union[Enum, str]]):
        return (a is None and b is None) or (str(a).lower() == str(b).lower())

    name = data.get('name', None)
    gsq_rm = gsq_rm_for_name(name)
    if gsq_rm is None:
        return None
    asset_class = data.get('assetClass', None)
    measure_type = data.get('measureType', None)
    if _enum_or_str_equal(asset_class, gsq_rm.asset_class) and _enum_or_str_equal(measure_type, gsq_rm.measure_type):
        result = copy.copy(gsq_rm)
        param = _decode_param(data)
        if param:
            result.parameters = param
        return result
    return None


def decode_risk_measure(data: Dict) -> RiskMeasure:
    result = _decode_gsq_risk_measure(data)
    if result is not None:
        return result
    if 'parameters' in data:
        result = ParameterisedRiskMeasure.from_dict(data)
        result.parameters = _decode_param(data)
    else:
        result = RiskMeasure.from_dict(data)
    return result


def decode_risk_measure_tuple(blob: Tuple[Dict, ...]) -> Tuple[RiskMeasure, ...]:
    return tuple(decode_risk_measure(s) for s in blob) if isinstance(blob, (tuple, list)) else None



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
from dataclasses import field, dataclass
from enum import Enum
from typing import Union, Tuple, Optional

import gs_quant.base as base
from gs_quant.base import handle_camel_case_args, Base, EnumBase
from gs_quant.instrument.core import Security


class TestEnum(EnumBase, Enum):
    Enum_1 = 'Enum_1'
    Enum_2 = 'Enum_2'


@handle_camel_case_args
@dataclass
class BaseSubclass(Base):
    instance_attr: str = field(default=None)
    attr_1: Optional[str] = field(default=None)
    attr_2: Tuple[Union[float, str], ...] = field(default=())
    attr_3: Union[Tuple[Optional[str], ...], str] = field(default=None)
    attr_4: Tuple[int] = field(default=None)
    attr_5: Tuple[int, str] = field(default=None)
    attr_6: TestEnum = field(default=None)


def test_handle_camel_case_args():
    # Handling camelcase args on init
    obj = BaseSubclass(instanceAttr="test")
    assert obj.instance_attr == "test"


def test_base_getter():
    obj = BaseSubclass(instance_attr="test")

    # Handling camelcase getter
    assert obj.instanceAttr == "test"


def test_base_setter():
    obj = BaseSubclass()

    # Handling camelcase setter
    obj.instanceAttr = "test"
    assert obj.instance_attr == "test"


def test_setter_coercion():
    base._is_supported_generic_cache = {}
    obj = BaseSubclass(instance_attr='test', attr_1=None, attr_2=('test', 1.0, 1), attr_3='test',
                       attr_4=(1,), attr_5=(3, 'test'), attr_6=TestEnum.Enum_1)
    obj.attr_1 = 'test'
    # str with type hint str is unchanged
    obj.attr_1 = 'Enum_1'
    assert isinstance(obj.attr_1, str)
    # Enum with type hint str is unchanged
    obj.attr_1 = TestEnum.Enum_1
    assert isinstance(obj.attr_1, TestEnum)
    obj.attr_2 = (1, 2, 3)
    obj.attr_3 = (None, 'test', None)
    obj.attr_4 = (1, 1, 1)
    obj.attr_5 = (0, 'test')
    # Enum with type hint Enum is unchanged
    obj.attr_6 = TestEnum.Enum_2
    assert isinstance(obj.attr_6, TestEnum)
    # all handled as type matches, we do not get to the generic coercion/default case
    assert not base._is_supported_generic_cache
    # str with type hint Enum gets cast to Enum
    obj.attr_6 = 'Enum_1'
    assert isinstance(obj.attr_6, TestEnum)
    assert TestEnum in base._is_supported_generic_cache


def test_security_from_dict():
    # Input dictionary
    input_dict = {
        '$type': 'LegDefinition',
        'bbid': 'TYU5 Comdty',
        'calculationTime': 13379,
        'properties': {
            'buysell': 'Buy',
            'identifier': 'TYU5 Comdty',
            'identifiertype': 'BloombergID',
            'quantity': 1.0
        },
        'queueingTime': 157
    }

    # Construct Security object
    security = Security.from_dict(input_dict)

    # Assertions
    assert security.bbid == 'TYU5 Comdty'



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import sys
from importlib.metadata import version as get_lib_version, PackageNotFoundError

from ._version import get_versions

name = "gs_quant"
__version__ = get_versions()['version']


def get_environment_summary():
    """
    Returns a summary of the Python version and the versions of specific libraries.
    """
    libraries = ['gs_quant', 'gs_quant_internal', 'numpy', 'pandas']
    summary = {
        'Python Version': sys.version,
    }
    for lib in libraries:
        try:
            lib_version = get_lib_version(lib)
        except PackageNotFoundError:
            lib_version = 'Not Installed'
        summary[lib] = lib_version
    return summary


__summary__ = get_environment_summary()
del get_versions

version = __version__
summary = __summary__

# Set up PyXll, if available
try:
    from .xl_interface.instrument_generation import install_hook

    install_hook()
except ModuleNotFoundError:
    pass

# Jupyter needs nest_asyncio to avoid event loop issues
try:
    from IPython import get_ipython

    ipython = get_ipython()
    if ipython and 'IPKernelApp' in get_ipython().config:
        import nest_asyncio

        nest_asyncio.apply()
except ImportError:
    pass

# Setup tracing for Jupyter
try:
    from gs_quant.tracing import tracing  # pylint: disable=unused-import
except ImportError:
    pass



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import math
from enum import auto, Enum
from typing import Dict, Union

import numpy as np
import pandas as pd

from gs_quant.api.gs.risk_models import GsFactorRiskModelApi, RiskModelDataMeasure, RiskModelDataAssetsRequest, \
    IntradayFactorDataSource
from gs_quant.data.core import DataContext
from gs_quant.datetime import date, time
from gs_quant.models.risk_model_utils import get_covariance_matrix_dataframe, build_factor_volatility_dataframe, \
    build_factor_data_map, build_pfp_data_dataframe
from gs_quant.target.risk_models import RiskModelUniverseIdentifierRequest


class ReturnFormat(Enum):
    """Alternative format for data to be returned from get_data functions"""
    JSON = auto()
    DATA_FRAME = auto()


class Factor:

    def __init__(self,
                 risk_model_id: str,
                 id_: str,
                 type_: str,
                 name: str = None,
                 category: str = None,
                 tooltip: str = None,
                 description: str = None,
                 glossary_description: str = None):
        self.__risk_model_id = risk_model_id
        self.__id = id_
        self.__name = name
        self.__type = type_
        self.__category = category
        self.__tooltip = tooltip
        self.__description = description
        self.__glossary_description = glossary_description

    @property
    def id(self):
        return self.__id

    @property
    def name(self):
        return self.__name

    @property
    def type(self) -> str:
        return self.__type

    @property
    def category(self) -> str:
        return self.__category

    @property
    def tooltip(self):
        return self.__tooltip

    @property
    def description(self):
        return self.__description

    @property
    def glossary_description(self):
        return self.__glossary_description

    @property
    def risk_model_id(self):
        return self.__risk_model_id

    def covariance(self,
                   factor: 'Factor',
                   start_date: date = DataContext.current.start_date,
                   end_date: date = DataContext.current.end_date,
                   format: ReturnFormat = ReturnFormat.DATA_FRAME) -> Union[Dict, pd.DataFrame]:
        """ Retrieve a Dataframe or Dictionary of date->covariance values between this factor and another for a date
                range """
        covariance_data_raw = GsFactorRiskModelApi.get_risk_model_data(self.risk_model_id,
                                                                       start_date,
                                                                       end_date,
                                                                       measures=[RiskModelDataMeasure.Covariance_Matrix,
                                                                                 RiskModelDataMeasure.Factor_Name,
                                                                                 RiskModelDataMeasure.Factor_Id],
                                                                       factors=list({self.name, factor.name}),
                                                                       limit_factors=False).get('results')

        covariance_data_raw = get_covariance_matrix_dataframe(covariance_data_raw)

        covariance_data_df = covariance_data_raw.stack().loc[pd.IndexSlice[:, self.name, factor.name]] * 252

        if format == ReturnFormat.JSON:
            return covariance_data_df.to_dict()

        return covariance_data_df.to_frame(name="covariance")

    def variance(self,
                 start_date: date = DataContext.current.start_date,
                 end_date: date = DataContext.current.end_date,
                 format: ReturnFormat = ReturnFormat.DATA_FRAME) -> Union[Dict, pd.DataFrame]:
        """ Retrieve a Dataframe or Dictionary of date->variance values for a factor over a date range """
        variance_raw_data = self.covariance(self, start_date, end_date, ReturnFormat.DATA_FRAME) \
            .rename(columns={"covariance": "variance"})

        if format == ReturnFormat.JSON:
            return variance_raw_data.squeeze().to_dict()

        return variance_raw_data

    def volatility(self,
                   start_date: date = DataContext.current.start_date,
                   end_date: date = DataContext.current.end_date,
                   format: ReturnFormat = ReturnFormat.DATA_FRAME) -> Union[Dict, pd.DataFrame]:
        """ Retrieve a Dataframe or Dictionary of date->volatility values for a factor over a date range """
        volatility_raw_data = GsFactorRiskModelApi.get_risk_model_data(self.risk_model_id,
                                                                       start_date,
                                                                       end_date,
                                                                       measures=[RiskModelDataMeasure.Factor_Volatility,
                                                                                 RiskModelDataMeasure.Factor_Id,
                                                                                 RiskModelDataMeasure.Factor_Name],
                                                                       factors=[self.name],
                                                                       limit_factors=False).get('results')

        volatility_data_df = build_factor_volatility_dataframe(volatility_raw_data, True, None) * math.sqrt(252)
        if format == ReturnFormat.JSON:
            return volatility_data_df.squeeze(axis=1).to_dict()

        return volatility_data_df

    def correlation(self,
                    other_factor,
                    start_date: date = DataContext.current.start_date,
                    end_date: date = DataContext.current.end_date,
                    format: ReturnFormat = ReturnFormat.DATA_FRAME) -> Union[Dict, pd.DataFrame]:
        """ Retrieve a Dataframe or Dictionary of date->correlation values between this factor and another for a date
        range """

        raw_data = GsFactorRiskModelApi.get_risk_model_data(self.risk_model_id,
                                                            start_date,
                                                            end_date,
                                                            measures=[RiskModelDataMeasure.Covariance_Matrix,
                                                                      RiskModelDataMeasure.Factor_Name,
                                                                      RiskModelDataMeasure.Factor_Id],
                                                            factors=[self.name, other_factor.name],
                                                            limit_factors=False).get('results')

        covariance_data_raw = get_covariance_matrix_dataframe(raw_data) * 252
        numerator = covariance_data_raw.stack().loc[pd.IndexSlice[:, self.name, other_factor.name]]
        denominator = np.sqrt(covariance_data_raw.stack().loc[pd.IndexSlice[:, self.name, self.name]] *
                              covariance_data_raw.stack().loc[pd.IndexSlice[:, other_factor.name, other_factor.name]])

        correlation_df = numerator / denominator

        if format == ReturnFormat.JSON:
            return correlation_df.to_dict()
        return correlation_df.to_frame(name="correlation")

    def returns(self,
                start_date: date = DataContext.current.start_date,
                end_date: date = DataContext.current.end_date,
                format: ReturnFormat = ReturnFormat.DATA_FRAME) -> Union[Dict, pd.DataFrame]:
        """ Retrieve a Dataframe or Dictionary of date->factor return values for a date range """
        factor_returns_raw = GsFactorRiskModelApi.get_risk_model_data(self.risk_model_id,
                                                                      start_date,
                                                                      end_date,
                                                                      measures=[RiskModelDataMeasure.Factor_Return,
                                                                                RiskModelDataMeasure.Factor_Name,
                                                                                RiskModelDataMeasure.Factor_Id],
                                                                      factors=[self.name],
                                                                      limit_factors=False).get('results')

        factor_returns_formatted = build_factor_data_map(factor_returns_raw, 'factorName', self.risk_model_id,
                                                         RiskModelDataMeasure.Factor_Return, None)
        factor_returns_df = factor_returns_formatted.rename(columns={self.name: 'return'}).rename_axis(None, axis=1)

        if format == ReturnFormat.JSON:
            return factor_returns_df.squeeze().to_dict()

        return factor_returns_df

    def intraday_returns(self,
                         start_time: time = DataContext.current.start_time,
                         end_time: time = DataContext.current.end_time,
                         data_source: Union[IntradayFactorDataSource, str] = IntradayFactorDataSource.GS_FMP,
                         format: ReturnFormat = ReturnFormat.DATA_FRAME) -> Union[Dict, pd.DataFrame]:
        """ Retrieve a Dataframe or Dictionary of timestamp->factor return values for a time range """

        max_interval = dt.timedelta(hours=23, minutes=59, seconds=59)
        current_start = start_time
        all_data = []

        while current_start < end_time:
            current_end = min(current_start + max_interval, end_time)
            factor_returns_batch = GsFactorRiskModelApi.get_risk_model_factor_data_intraday(self.risk_model_id,
                                                                                            current_start,
                                                                                            current_end,
                                                                                            data_source=data_source,
                                                                                            factors=[self.name])
            all_data.extend(factor_returns_batch)
            current_start = current_end + dt.timedelta(seconds=1)  # Move to the next interval

        factor_returns_df = pd.DataFrame(all_data)
        try:
            factor_returns_df.set_index(['time'], inplace=True)
        except KeyError:
            factor_returns_df = pd.DataFrame()

        factor_returns_df.drop(columns=["factorCategory", "factor", "factorId"], errors='ignore', inplace=True)

        if format == ReturnFormat.JSON:
            return factor_returns_df.squeeze().to_dict()

        return factor_returns_df

    def mimicking_portfolio(self,
                            start_date: date = DataContext.current.start_date,
                            end_date: date = DataContext.current.end_date,
                            assets: RiskModelDataAssetsRequest = RiskModelDataAssetsRequest(
                                identifier=RiskModelUniverseIdentifierRequest.bbid, universe=()),
                            format: ReturnFormat = ReturnFormat.DATA_FRAME):
        """ Retrieves a timeseries of factor mimicking portfolios for a date range."""

        results = GsFactorRiskModelApi.get_risk_model_data(self.risk_model_id,
                                                           start_date,
                                                           end_date,
                                                           assets=assets,
                                                           measures=[RiskModelDataMeasure.Factor_Portfolios,
                                                                     RiskModelDataMeasure.Factor_Id,
                                                                     RiskModelDataMeasure.Factor_Name],
                                                           factors=[self.name],
                                                           limit_factors=False).get('results')

        factor_portfolio_df = build_pfp_data_dataframe(results, True)

        factor_portfolio_df = factor_portfolio_df.reset_index()
        factor_portfolio_df = factor_portfolio_df.pivot(index="date",
                                                        columns="identifier",
                                                        values=self.name)

        if format.value == ReturnFormat.JSON.value:
            return factor_portfolio_df.to_dict(orient='index')
        return factor_portfolio_df
